"""Submodule providing edge prediction models based on Sklearn Models."""
from .sklearn_model_edge_prediction_adapter import SklearnModelEdgePredictionAdapter

__all__ = [
    "SklearnModelEdgePredictionAdapter"
]


"""Submodule with utilities relative to GPU availability."""
from typing import List
import shutil
import warnings


def get_available_gpus() -> List[str]:
    """Return list with IDs of available GPU devices."""
    try:
        import tensorflow as tf
        return tf.config.experimental.list_physical_devices('GPU')
    except ModuleNotFoundError:
        return []


def command_is_available(command_name: str) -> bool:
    """Return whether given bash command is available in PATH.

    Parameters
    ------------------
    command_name: str,
        The command to check availability for.

    Returns
    ------------------
    Boolean representing if the command is available in PATH.
    """
    return shutil.which(command_name) is not None


def has_nvidia_drivers() -> bool:
    """Return whether NVIDIA drivers can be detected."""
    return command_is_available("nvidia-smi")


def has_rocm_drivers() -> bool:
    """Return whether ROCM drivers can be detected."""
    return command_is_available("rocm-smi")


def get_available_gpus_number() -> int:
    """Return whether GPUs can be detected."""
    return len(get_available_gpus())


def has_single_gpu() -> bool:
    """Return whether there is only a GPU available."""
    return get_available_gpus_number() == 1


def has_gpus() -> bool:
    """Return whether GPUs can be detected."""
    return get_available_gpus_number() > 0


def execute_gpu_checks():
    """Executes the GPU checks and raises the proper warnings."""
    # To avoid some nighmares we check availability of GPUs.
    if not has_gpus():
        # We check for drivers to try and give a more explainatory
        # warning about the absence of GPUs.
        if has_nvidia_drivers():
            warnings.warn(
                "It was not possible to detect GPUs but the system "
                "has NVIDIA drivers installed.\n"
                "It is very likely there is some mis-configuration "
                "with your TensorFlow instance.\n"
                "The model will train a LOT faster if you figure "
                "out what may be the cause of this issue on your "
                "system: sometimes a simple reboot will do a lot of good.\n"
                "If you are currently on COLAB, remember to enable require "
                "a GPU instance from the menu!"
            )
        elif has_rocm_drivers():
            warnings.warn(
                "It was not possible to detect GPUs but the system "
                "has ROCM drivers installed.\n"
                "It is very likely there is some mis-configuration "
                "with your TensorFlow instance.\n"
                "The model will train a LOT faster if you figure "
                "out what may be the cause of this issue on your "
                "system: sometimes a simple reboot will do a lot of good."
            )
        else:
            warnings.warn(
                "It was neither possible to detect GPUs nor GPU drivers "
                "of any kind on your system (neither CUDA or ROCM).\n"
                "The model will proceed with trainining, but it will be "
                "significantly slower than what would be possible "
                "with GPU acceleration."
            )


"""Class for creating a Perceptron model for edge prediction tasks."""
from typing import Union
import pandas as pd
import numpy as np
from ensmallen import Graph
from .feed_forward_neural_network import FeedForwardNeuralNetwork


class Perceptron(FeedForwardNeuralNetwork):

    def __init__(
        self,
        graph: Graph,
        embedding_size: int = None,
        embedding: Union[np.ndarray, pd.DataFrame] = None,
        edge_embedding_method: str = "Concatenate",
        optimizer: str = "nadam",
        trainable_embedding: bool = False,
        use_dropout: bool = True,
        dropout_rate: float = 0.5,
        use_edge_metrics: bool = False,
        task_name: str = "EDGE_PREDICTION"
    ):
        """Create new edge prediction Perceptron model object.

        Parameters
        --------------------
        graph: Graph,
            The graph object to base the model on.
        embedding_size: int = None,
            Dimension of the embedding.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding: Union[np.ndarray, pd.DataFrame] = None,
            The seed embedding to be used.
            Note that it is not possible to provide at once both
            the embedding and either the vocabulary size or the embedding size.
        edge_embedding_method: str = "Concatenate",
            Method to use to create the edge embedding.
        optimizer: str = "nadam",
            Optimizer to use during the training.
        trainable_embedding: bool = False,
            Whether to allow for trainable embedding.
        use_dropout: bool = True,
            Whether to use dropout.
        dropout_rate: float = 0.5,
            Dropout rate.
        use_edge_metrics: bool = False,
            Whether to return the edge metrics.
        task_name: str = "EDGE_PREDICTION",
            The name of the task to build the model for.
            The currently supported task names are `EDGE_PREDICTION` and `EDGE_LABEL_PREDICTION`.
            The default task name is `EDGE_PREDICTION`.
        """
        super().__init__(
            graph,
            embedding_size=embedding_size,
            embedding=embedding,
            edge_embedding_method=edge_embedding_method,
            optimizer=optimizer,
            trainable_embedding=trainable_embedding,
            use_dropout=use_dropout,
            dropout_rate=dropout_rate,
            use_edge_metrics=use_edge_metrics,
            task_name=task_name
        )



"""Submodule with utils for interface with Sklearn models."""
import functools
from typing import Dict, Type
from sklearn.base import ClassifierMixin
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from userinput.utils import closest
import numpy as np
from sanitize_ml_labels import sanitize_ml_labels
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    average_precision_score,
    precision_score,
    recall_score,
    roc_auc_score
)


def get_default_decision_tree_classifier(**kwargs: Dict) -> DecisionTreeClassifier:
    """Returns a default decision tree classifier.

    On the parameters used
    ----------------------------
    The parameters used for this model are reasonable
    general use parameters, but if you had any better
    general parameters please do feel free to drop an
    issue on the Embiggen repository with your proposal.
    Your insight is absolutely welcomed.
    """
    return DecisionTreeClassifier(**{
        **dict(
            max_depth=10,
            random_state=42,
            class_weight="balanced"
        ),
        **kwargs
    })


def get_default_extra_tree_classifier(**kwargs: Dict) -> ExtraTreesClassifier:
    """Returns a default extra tree classifier.

    On the parameters used
    ----------------------------
    The parameters used for this model are reasonable
    general use parameters, but if you had any better
    general parameters please do feel free to drop an
    issue on the Embiggen repository with your proposal.
    Your insight is absolutely welcomed.
    """
    return ExtraTreesClassifier(**{
        **dict(
            n_estimators=500,
            max_depth=10,
            random_state=42,
            bootstrap=True,
            n_jobs=-1,
            max_features="sqrt",
            class_weight="balanced"
        ),
        **kwargs
    })


def get_default_random_forest_classifier(**kwargs: Dict) -> RandomForestClassifier:
    """Returns a default random forest classifier.

    On the parameters used
    ----------------------------
    The parameters used for this model are reasonable
    general use parameters, but if you had any better
    general parameters please do feel free to drop an
    issue on the Embiggen repository with your proposal.
    Your insight is absolutely welcomed.
    """
    return RandomForestClassifier(**{
        **dict(
            n_estimators=500,
            max_depth=10,
            random_state=42,
            bootstrap=True,
            n_jobs=-1,
            max_features="sqrt",
            class_weight="balanced"
        ),
        **kwargs
    })


def get_default_k_neighbours_classifier(**kwargs: Dict) -> KNeighborsClassifier:
    """Returns a default K-neighbours classifier.

    On the parameters used
    ----------------------------
    The parameters used for this model are reasonable
    general use parameters, but if you had any better
    general parameters please do feel free to drop an
    issue on the Embiggen repository with your proposal.
    Your insight is absolutely welcomed.
    """
    return KNeighborsClassifier(**{
        **dict(
            n_neighbors=10,
            weights="distance",
            n_jobs=-1,
        ),
        **kwargs
    })


def get_default_mlp_classifier(**kwargs: Dict) -> MLPClassifier:
    """Returns a default MLP classifier.

    On the parameters used
    ----------------------------
    The parameters used for this model are reasonable
    general use parameters, but if you had any better
    general parameters please do feel free to drop an
    issue on the Embiggen repository with your proposal.
    Your insight is absolutely welcomed.
    """
    return MLPClassifier(**{
        **dict(
            activation="relu",
            hidden_layer_sizes=100,
            solver="adam",
            max_iter=500,
            random_state=42,
            early_stopping=True,
        ),
        **kwargs
    })


available_default_sklearn_classifiers = {
    "DecisionTreeClassifier": get_default_decision_tree_classifier,
    "ExtraTreesClassifier": get_default_extra_tree_classifier,
    "RandomForestClassifier": get_default_random_forest_classifier,
    "KNeighborsClassifier": get_default_k_neighbours_classifier,
    "MLPClassifier": get_default_mlp_classifier
}


def is_default_sklearn_classifier(model_name: str) -> bool:
    """Returns whether we have default models for the provided model name.

    Parameters
    ------------------------
    model_name: str
        Name of the default model to check for.
    """
    return model_name.lower() in {
        key.lower()
        for key in available_default_sklearn_classifiers
    }


def must_be_default_sklearn_classifier(model_name: str) -> bool:
    """Raises exception if we do not have default models for the provided model name.

    Parameters
    ------------------------
    model_name: str
        Name of the default model to check for.

    Raises
    ------------------------
    ValueError
        If the provided model is not currently available as a default.
    """
    if not is_default_sklearn_classifier(model_name):
        raise ValueError(
            (
                "The provided model name `{model_name}` is not a supported sklearn "
                "classifier among the default available ones. Possibly you meant {candidate}?"
            ).format(
                model_name=model_name,
                candidate=closest(
                    model_name, available_default_sklearn_classifiers.keys())
            )
        )


def get_sklearn_default_classifier(model_name: str, **kwargs: Dict):
    """Returns instance of default Sklearn classifier.

    Parameters
    ------------------------
    model_name: str
        Name of the default model to load.
    **kwargs: Dict
        Arguments to forward to the sklearn model construction.
    """
    must_be_default_sklearn_classifier(model_name)
    return {
        key.lower(): callback
        for key, callback in available_default_sklearn_classifiers.items()
    }[model_name.lower()](**kwargs)


def is_sklearn_classifier_model(candidate_model) -> bool:
    """Returns whether a given object is a Sklearn classifier model."""
    return issubclass(candidate_model.__class__, ClassifierMixin) and all(
        hasattr(candidate_model, method_name)
        for method_name in (
            "predict_proba",
            "predict",
            "fit",
        )
    )


def must_be_an_sklearn_classifier_model(candidate_model):
    """Raises an exception if the provided object is not an sklearn classifier model."""
    if not is_sklearn_classifier_model(candidate_model):
        raise ValueError(
            (
                "The provided object of type {} is not a valid sklearn classifier model "
                "that can be adapted for this class as it is not a subclass of `ClassifierMixin`."
            ).format(type(candidate_model))
        )


def evaluate_sklearn_classifier(
    classifier: Type[ClassifierMixin],
    X: np.ndarray,
    y: np.ndarray,
    multiclass_or_multilabel: bool,
    **kwargs: Dict
) -> Dict[str, float]:
    """Return dict with evaluation of the model on classification task.

    Parameters
    --------------
    classifier: Type[ClassifierMixin]
        The classifier to evaluate.
    X: np.ndarray
        The input data.
    y: np.ndarray
        The output labels
    multiclass_or_multilabel: bool
        Whether the provided labels are multiclass or multilabel,
        that is the model needs to compute more output
        classes for a given input.
    **kwargs: Dict
        Aguments to forward to the model predict proba method.
    """

    must_be_an_sklearn_classifier_model(classifier)

    predictions = classifier.predict_proba(X, **kwargs)

    # If this is a binary prediction, we need to handle
    # this as a corner case.
    if not multiclass_or_multilabel:
        predictions = predictions[:, 1]

    integer_predictions = classifier.predict(X, **kwargs)

    # Depending on whether this task is binary or multiclall/multilabel
    # we need to set the average parameter for the metrics.
    if multiclass_or_multilabel:
        average_methods = average_methods_probs = ("macro", "weighted")
    else:
        average_methods = ("binary",)
        average_methods_probs = ("micro", "macro", "weighted")

    @functools.wraps(roc_auc_score)
    def wrapper_roc_auc_score(*args, **kwargs):
        return roc_auc_score(*args, **kwargs, multi_class="ovr")

    return {
        **{
            sanitize_ml_labels(integer_metric.__name__): integer_metric(y, integer_predictions)
            for integer_metric in (
                accuracy_score,
                balanced_accuracy_score,
            )
        },
        **{
            "{}{}".format(
                sanitize_ml_labels(averaged_integer_metric.__name__),
                " ".format(
                    average_method) if average_method != "binary" else ""
            ): averaged_integer_metric(y, integer_predictions, average=average_method)
            for averaged_integer_metric in (
                f1_score,
                precision_score,
                recall_score,
            )
            for average_method in average_methods
        },
        **{
            "{} {}".format(
                sanitize_ml_labels(probabilistic_metric.__name__),
                average_method
            ): probabilistic_metric(
                y,
                predictions,
                average=average_method
            )
            for probabilistic_metric in (
                # AUPRC in sklearn is only supported for binary labels
                *((average_precision_score,)
                  if not multiclass_or_multilabel else ()),
                wrapper_roc_auc_score
            )
            for average_method in average_methods_probs
        },
    }



"""Submodule providing node-label prediction models based on Sklearn Models."""

from .sklearn_model_node_label_prediction_adapter import SklearnModelNodeLabelPredictionAdapter

__all__ = [
    "SklearnModelNodeLabelPredictionAdapter"
]



"""Submodule providing edge-label prediction models based on Sklearn Models."""

from .sklearn_model_edge_label_prediction_adapter import SklearnModelEdgeLabelPredictionAdapter

__all__ = [
    "SklearnModelEdgeLabelPredictionAdapter"
]


"""Keras Sequence for running GNN on graph edge prediction."""
from typing import Tuple, Union, List, Optional

import numpy as np
import pandas as pd
import tensorflow as tf
from ensmallen import Graph  # pylint: disable=no-name-in-module
from ..utils import tensorflow_version_is_higher_or_equal_than
from .edge_prediction_sequence import EdgePredictionSequence


class GNNEdgePredictionSequence(EdgePredictionSequence):
    """Keras Sequence for running GNN on graph edge prediction."""

    def __init__(
        self,
        graph: Graph,
        node_features: Optional[Union[pd.DataFrame, List[pd.DataFrame], np.ndarray, List[np.ndarray]]] = None,
        use_node_types: bool = False,
        use_edge_metrics: bool = False,
        return_node_ids: bool = True,
        batch_size: int = 2**15,
        negative_samples_rate: float = 0.9,
        avoid_false_negatives: bool = False,
        graph_to_avoid: Optional[Graph] = None,
        batches_per_epoch: Union[int, str] = "auto",
        elapsed_epochs: int = 0,
        random_state: int = 42
    ):
        """Create new GNNEdgePredictionSequence object.

        Parameters
        --------------------------------
        graph: Graph,
            The graph from which to sample the edges.
        use_node_types: bool = False,
            Whether to return the node types.
        use_edge_types: bool = False,
            Whether to return the edge types.
        use_edge_metrics: bool = False,
            Whether to return the edge metrics.
        return_node_ids: bool = True,
            Whether to return the node IDs.
        batch_size: int = 2**15,
            The batch size to use.
        negative_samples_rate: float = 0.9,
            Factor of negatives to use in every batch.
            For example, with a batch size of 128 and negative_samples_rate equal
            to 0.5, there will be 64 positives and 64 negatives.
        avoid_false_negatives: bool = False,
            Whether to filter out false negatives.
            By default False.
            Enabling this will slow down the batch generation while (likely) not
            introducing any significant gain to the model performance.
        graph_to_avoid: Graph = None,
            Graph to avoid when generating the edges.
            This can be the validation component of the graph, for example.
            More information to how to generate the holdouts is available
            in the Graph package.
        batches_per_epoch: Union[int, str] = "auto",
            Number of batches per epoch.
            If auto, it is used: `10 * edges number /  batch size`
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        random_state: int = 42,
            The random_state to use to make extraction reproducible.
        """
        self._node_features = [
            node_feature.values
            if isinstance(node_feature, pd.DataFrame)
            else node_feature
            for node_feature in node_features
        ]
        self._return_node_ids = return_node_ids
        self._current_index = 0
        super().__init__(
            graph,
            use_node_types=use_node_types,
            use_edge_types=False,
            use_edge_metrics=use_edge_metrics,
            batch_size=batch_size,
            negative_samples_rate=negative_samples_rate,
            avoid_false_negatives=avoid_false_negatives,
            filter_none_values=False,
            graph_to_avoid=graph_to_avoid,
            batches_per_epoch=batches_per_epoch,
            elapsed_epochs=elapsed_epochs,
            random_state=random_state,
        )

    def __call__(self):
        """Return next batch using an infinite generator model."""
        self._current_index += 1
        return self[self._current_index]

    def into_dataset(self) -> tf.data.Dataset:
        """Return dataset generated out of the current sequence instance.

        Implementative details
        ---------------------------------
        This method handles the conversion of this Keras Sequence into
        a TensorFlow dataset, also handling the proper dispatching according 
        to what version of TensorFlow is installed in this system.

        Returns
        ----------------------------------
        Dataset to be used for the training of a model
        """

        #################################################################
        # Handling kernel creation when TensorFlow is a modern version. #
        #################################################################

        if tensorflow_version_is_higher_or_equal_than("2.5.0"):
            node_shapes = [
                *[
                    tf.TensorSpec(
                        shape=(self._batch_size, node_features.shape[1]),
                        dtype=node_features.dtype
                    )
                    for node_features in self._node_features
                ],
                * (
                    (tf.TensorSpec(shape=(self._batch_size, ), dtype=tf.int32),)
                    if self._return_node_ids else ()
                ),
                * (
                    (tf.TensorSpec(
                        shape=(self._batch_size, self._graph.get_maximum_multilabel_count()), dtype=tf.int32),)
                    if self._use_node_types else ()
                )
            ]
            return tf.data.Dataset.from_generator(
                self,
                output_signature=(
                    (
                        # Node shapes relative to the source node
                        *node_shapes,
                        # Node shapes relative to the destination node
                        *node_shapes
                    ),
                    tf.TensorSpec(
                        shape=(self._batch_size,),
                        dtype=tf.bool
                    )
                )
            )

        output_node_data_types = [
            *[
                node_features.dtype
                for node_features in self._node_features
            ],
            * (
                (tf.int32,)
                if self._return_node_ids else ()
            ),
            * (
                (tf.int32,)
                if self._use_node_types else ()
            )
        ]

        output_node_shapes = [
            *[
                tf.TensorShape(
                    [self._batch_size, node_features.shape[1]]
                )
                for node_features in self._node_features
            ],
            * (
                (tf.TensorShape([self._batch_size, ]),)
                if self._return_node_ids else ()
            ),
            * (
                (tf.TensorShape(
                    [self._batch_size, self._graph.get_maximum_multilabel_count()]),)
                if self._use_node_types else ()
            )
        ]

        return tf.data.Dataset.from_generator(
            self,
            output_types=(
                (
                    # The source nodes
                    *output_node_data_types,
                    # The destination nodes
                    *output_node_data_types
                ),
                tf.bool
            ),
            output_shapes=(
                (
                    # The source nodes shapes
                    *output_node_shapes,
                    # The destination nodes shapes
                    *output_node_shapes,
                ),
                tf.TensorShape([self._batch_size, ]),
            )
        )

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        (sources, source_node_types, destinations,
         destination_node_types, _, _), labels = super().__getitem__(idx)[0]

        source_ids = None
        if self._return_node_ids:
            source_ids = sources

        destination_ids = None
        if self._return_node_ids:
            destination_ids = destinations

        return ((
            tuple([
                node_feature[sources]
                for node_feature in self._node_features
            ] + [
                value
                for value in (
                    source_ids, source_node_types
                )
                if value is not None
            ] + [
                node_feature[destinations]
                for node_feature in self._node_features
            ] + [
                value
                for value in (
                    destination_ids, destination_node_types
                )
                if value is not None
            ]),
            labels
        ),)



"""Module providing adapter class making edge prediction possible in sklearn models."""
from sklearn.base import ClassifierMixin
from typing import Type, List, Dict, Union, Optional, Tuple
import pandas as pd
import numpy as np
from ensmallen import Graph
from ...utils import must_be_an_sklearn_classifier_model, evaluate_sklearn_classifier
from ...transformers import EdgePredictionTransformer, GraphTransformer


class SklearnModelEdgePredictionAdapter:
    """Class wrapping Sklearn models for running ."""

    def __init__(
        self,
        model_instance: Type[ClassifierMixin],
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: Type[ClassifierMixin]
            The class instance to be adapted into edge prediction.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        must_be_an_sklearn_classifier_model(model_instance)
        self._model_instance = model_instance
        # We want to mask the decorator class name
        self.__class__.__name__ = model_instance.__class__.__name__
        self.__class__.__doc__ = model_instance.__class__.__doc__

    @staticmethod
    def _trasform_graphs_into_edge_embedding(
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the training of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the sum of the directed edges in the positive and
            negative graphs. In this matrix, first we expect the
            positive graph edge features and secondly the negative graph
            edge features. We will be shuffling the edge features
            alongside the edge embedding to have everything aligned
            correctly.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and training.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        lpt = EdgePredictionTransformer(
            method=edge_embedding_method,
            aligned_node_mapping=aligned_node_mapping
        )

        lpt.fit(node_features)

        return lpt.transform(
            positive_graph=positive_graph,
            negative_graph=negative_graph,
            edge_features=edge_features,
            random_state=random_state
        )

    @staticmethod
    def _trasform_graph_into_edge_embedding(
        graph: Union[Graph, List[List[str]], List[List[int]]],
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
    ) -> np.ndarray:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the training of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the sum of the directed edges in the positive and
            negative graphs. In this matrix, first we expect the
            positive graph edge features and secondly the negative graph
            edge features. We will be shuffling the edge features
            alongside the edge embedding to have everything aligned
            correctly.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        gt = GraphTransformer(
            method=edge_embedding_method,
            aligned_node_mapping=aligned_node_mapping
        )

        gt.fit(node_features)

        return gt.transform(
            graph=graph,
            edge_features=edge_features
        )

    def fit(
        self,
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
        **kwargs
    ):
        """Execute fitting of the model.

        Parameters
        ------------------
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the training of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the sum of the directed edges in the positive and
            negative graphs. In this matrix, first we expect the
            positive graph edge features and secondly the negative graph
            edge features. We will be shuffling the edge features
            alongside the edge embedding to have everything aligned
            correctly.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and training.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model fitting.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        self._model_instance.fit(
            *SklearnModelEdgePredictionAdapter._trasform_graphs_into_edge_embedding(
                positive_graph=positive_graph,
                negative_graph=negative_graph,
                node_features=node_features,
                edge_features=edge_features,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
                random_state=random_state,
            ),
            **kwargs
        )

    def evaluate(
        self,
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge prediction task on the provided data.

        Parameters
        ------------------
        positive_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        negative_graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the evaluation of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the sum of the directed edges in the positive and
            negative graphs. In this matrix, first we expect the
            positive graph edge features and secondly the negative graph
            edge features. We will be shuffling the edge features
            alongside the edge embedding to have everything aligned
            correctly.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and evaluation.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return evaluate_sklearn_classifier(
            self._model_instance,
            *SklearnModelEdgePredictionAdapter._trasform_graphs_into_edge_embedding(
                positive_graph=positive_graph,
                negative_graph=negative_graph,
                node_features=node_features,
                edge_features=edge_features,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
                random_state=random_state,
            ),
            multiclass_or_multilabel=False,
            **kwargs
        )

    def predict_proba(
        self,
        graph: Union[Graph, List[List[str]], List[List[int]]],
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge prediction task on the provided data.

        Parameters
        ------------------
        graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the evaluation of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict_proba(
            SklearnModelEdgePredictionAdapter._trasform_graph_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                edge_features=edge_features,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
            ),
            **kwargs
        )

    def predict(
        self,
        graph: Union[Graph, List[List[str]], List[List[int]]],
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge prediction task on the provided data.

        Parameters
        ------------------
        graph: Union[Graph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the evaluation of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Warns
        ------------------
        If the node features are provided as a numpy array it will not be possible
        to check whether the nodes in the graphs are aligned with the features.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict(
            SklearnModelEdgePredictionAdapter._trasform_graph_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                edge_features=edge_features,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
            ),
            **kwargs
        )



"""Submodule providing models for edge-label prediction."""
from .node_label_prediction_sklearn import SklearnModelNodeLabelPredictionAdapter

__all__ = [
    "SklearnModelNodeLabelPredictionAdapter"
]



"""Class for an abstract edge prediction model."""
from multiprocessing.sharedctypes import Value
from typing import Dict, Union, Optional
import warnings
from embiggen.sequences.binary_edge_label_prediction_sequence import BinaryEdgeLabelPredictionSequence
from embiggen.sequences.edge_prediction_evaluation_sequence import EdgePredictionEvaluationSequence

import numpy as np
import pandas as pd
from ensmallen import Graph
from extra_keras_metrics import get_standard_binary_metrics, get_sparse_multiclass_metrics
from tensorflow.keras.layers import Layer, Input, Concatenate  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.models import Model  # pylint: disable=import-error,no-name-in-module
from userinput import set_validator
from userinput.utils import closest

from ...embedders.tensorflow_embedders import TensorFlowEmbedder
from ...sequences import EdgePredictionSequence, EdgeLabelPredictionSequence
from .layers import edge_embedding_layer

edge_prediction_supported_tasks = [
    "EDGE_PREDICTION",
    "EDGE_LABEL_PREDICTION",
    "BINARY_EDGE_LABEL_PREDICTION"
]


class EdgePredictionModel(TensorFlowEmbedder):

    def __init__(
        self,
        graph: Graph,
        embedding_size: int = None,
        embedding: Optional[Union[np.ndarray, pd.DataFrame]] = None,
        edge_embedding_method: str = "Concatenate",
        optimizer: str = "nadam",
        trainable_embedding: bool = False,
        use_dropout: bool = True,
        dropout_rate: float = 0.5,
        use_edge_metrics: bool = False,
        positive_edge_type: Optional[Union[str, int]] = None,
        task_name: str = "EDGE_PREDICTION"
    ):
        """Create new abstract Edge Prediction Model object.

        Parameters
        --------------------
        graph: Graph
            The graph object to base the model on.
        embedding_size: int = None
            Dimension of the embedding.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding: Union[np.ndarray, pd.DataFrame] = None
            The seed embedding to be used.
            Note that it is not possible to provide at once both
            the embedding and either the vocabulary size or the embedding size.
        edge_embedding_method: str = "Concatenate"
            Method to use to create the edge embedding.
        optimizer: str = "nadam"
            Optimizer to use during the training.
        trainable_embedding: bool = False
            Whether to allow for trainable embedding.
        use_dropout: bool = True
            Whether to use dropout.
        dropout_rate: float = 0.5
            Dropout rate.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        positive_edge_type: Optional[Union[str, int]] = None
            The type for the positive class.
            If None, we check if the provided graph has two classes,
            and if so we use the minority class as the positive class.
            If the graph does not have two classes, we will raise an error.
        task_name: str = "EDGE_PREDICTION"
            The name of the task to build the model for.
            The currently supported task names are `EDGE_PREDICTION` and `EDGE_LABEL_PREDICTION`.
            The default task name is `EDGE_PREDICTION`.
        """
        supported_edge_embedding_methods = list(edge_embedding_layer.keys())
        if not set_validator(supported_edge_embedding_methods)(edge_embedding_method):
            raise ValueError(
                (
                    "The provided edge embedding method name `{edge_embedding_method}` "
                    "is not supported. Did you mean {closest}?\n"
                    "The supported edge embedding method names are: {supported_tasks}."
                ).format(
                    edge_embedding_method=edge_embedding_method,
                    closest=closest(edge_embedding_method,
                                    supported_edge_embedding_methods),
                    supported_tasks=", ".join(
                        supported_edge_embedding_methods),
                )
            )
        if not set_validator(edge_prediction_supported_tasks)(task_name):
            raise ValueError(
                (
                    "The provided task name `{task_name}` is not supported. Did you mean {closest}?\n"
                    "The supported task names are: {supported_tasks}."
                ).format(
                    task_name=task_name,
                    closest=closest(
                        task_name, edge_prediction_supported_tasks),
                    supported_tasks=", ".join(edge_prediction_supported_tasks),
                )
            )
        if embedding is None and not trainable_embedding:
            warnings.warn(
                "The embedding was not provided, therefore a new random Normal embedding "
                "will be allocated, but also the trainable embedding flag was left set to false.\n"
                "This means that the embedding will not be trained, and the nodes will have "
                "exclusively random features associated to them."
            )

        if positive_edge_type is not None and task_name != "BINARY_EDGE_LABEL_PREDICTION":
            raise ValueError(
                "The parameter `positive_edge_type` was provided but the task "
                "provided in not a binary edge-label prediction"
            )

        self._positive_edge_type = positive_edge_type
        self._graph = graph
        self._task_name = task_name
        self._use_edge_metrics = use_edge_metrics
        self._edge_embedding_method = edge_embedding_method
        self._model_name = self.__class__.__name__
        self._use_dropout = use_dropout
        self._dropout_rate = dropout_rate
        super().__init__(
            vocabulary_size=self._graph.get_nodes_number(),
            embedding_size=embedding_size,
            optimizer=optimizer,
            embedding=embedding,
            trainable_embedding=trainable_embedding
        )

    def _compile_model(self) -> Model:
        """Compile model."""
        if self._task_name in ("EDGE_PREDICTION", "BINARY_EDGE_LABEL_PREDICTION"):
            self._model.compile(
                loss="binary_crossentropy",
                optimizer=self._optimizer,
                metrics=get_standard_binary_metrics(),
            )
        elif self._task_name == "EDGE_LABEL_PREDICTION":
            self._model.compile(
                loss="sparse_categorical_crossentropy",
                optimizer=self._optimizer,
                metrics=get_sparse_multiclass_metrics(),
            )
        else:
            raise ValueError("Unreacheable!")

    def _build_model_body(self, input_layer: Layer) -> Layer:
        """Build new model body for Edge prediction.

        Parameters
        --------------------
        input_layer: Layer,
            The previous layer to be used as input of the model.
        """
        raise NotImplementedError(
            "The method _build_model_body must be implemented in the child classes."
        )

    def _build_model(self) -> Model:
        """Build new model for Edge prediction."""
        embedding_layer = edge_embedding_layer[self._edge_embedding_method](
            nodes_number=self._graph.get_nodes_number(),
            embedding_size=self._embedding_size,
            embedding=self._embedding,
            use_dropout=self._use_dropout,
            dropout_rate=self._dropout_rate
        )

        inputs = [*embedding_layer.inputs]
        edge_embedding = embedding_layer(None)

        if self._use_edge_metrics:
            edge_metrics_input = Input(
                (self._graph.get_number_of_available_edge_metrics(),), name="EdgeMetrics")
            inputs.append(edge_metrics_input)
            edge_embedding = Concatenate()([
                edge_embedding,
                edge_metrics_input
            ])

        return Model(
            inputs=inputs,
            outputs=self._build_model_body(edge_embedding),
            name="{}_{}".format(
                self._model_name,
                self._edge_embedding_method
            )
        )

    def _validate_fit_parameters(
        self,
        valid_graph: Optional[Graph] = None,
        negative_valid_graph: Optional[Graph] = None,
    ):
        if negative_valid_graph is not None:
            if self._task_name != "EDGE_PREDICTION":
                raise ValueError(
                    "The negative validation graph was provided, but it is only used "
                    "in edge prediction tasks and this is not an edge prediction task. "
                )
            if valid_graph is None:
                raise ValueError(
                    "The negative validation graph was provided, but the validation graph "
                    "was not provided."
                )

    def fit(
        self,
        train_graph: Graph,
        valid_graph: Optional[Graph] = None,
        negative_valid_graph: Optional[Graph] = None,
        batch_size: int = 2**15,
        batches_per_epoch: Union[int, str] = "auto",
        negative_samples_rate: float = 0.75,
        epochs: int = 10000,
        early_stopping_monitor: str = "loss",
        early_stopping_min_delta: float = 0.001,
        early_stopping_patience: int = 10,
        early_stopping_mode: str = "min",
        reduce_lr_monitor: str = "loss",
        reduce_lr_min_delta: float = 0.01,
        reduce_lr_patience: int = 5,
        reduce_lr_mode: str = "min",
        reduce_lr_factor: float = 0.9,
        sample_only_edges_with_heterogeneous_node_types: bool = False,
        verbose: int = 2,
        ** kwargs: Dict
    ) -> pd.DataFrame:
        """Train model and return training history dataframe.

        Parameters
        -------------------
        train_graph: Graph
            Graph object to use for training.
        batch_size: int = 2**16,
            Batch size for the training process.
        batches_per_epoch: int = 2**10,
            Number of batches to train for in each epoch.
        negative_samples_rate: float = 0.5,
            Rate of unbalancing in the batch.
        epochs: int = 10000,
            Epochs to train the model for.
        early_stopping_monitor: str = "loss",
            Metric to monitor for early stopping. 
        early_stopping_min_delta: float = 0.001,
            Minimum delta of metric to stop the training.
        early_stopping_patience: int = 10,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which trigger early stopping.
        early_stopping_mode: str = "min",
            Direction of the variation of the monitored metric for early stopping.
        reduce_lr_monitor: str = "loss",
            Metric to monitor for reducing learning rate.
        reduce_lr_min_delta: float = 0.01,
            Minimum delta of metric to reduce learning rate.
        reduce_lr_patience: int = 5,
            Number of epochs to wait for when the given minimum delta is not
            achieved after which reducing learning rate.
        reduce_lr_mode: str = "min",
            Direction of the variation of the monitored metric for learning rate.
        reduce_lr_factor: float = 0.9,
            Factor for reduction of learning rate.
        sample_only_edges_with_heterogeneous_node_types: bool = False
            Whether to only sample edges between heterogeneous node types.
            This may be useful when training a model to predict between
            two portions in a bipartite graph.
        verbose: int = 2,
            Wethever to show the loading bar.
            Specifically, the options are:
            * 0 or False: No loading bar.
            * 1 or True: Showing only the loading bar for the epochs.
            * 2: Showing loading bar for both epochs and batches.
        ** kwargs: Dict,
            Parameteres to pass to the dit call.

        Returns
        --------------------
        Dataframe with traininhg history.
        """
        self._validate_fit_parameters(
            negative_valid_graph=negative_valid_graph,
            valid_graph=valid_graph
        )

        validation_sequence = None
        if self._task_name == "EDGE_PREDICTION":
            training_sequence = EdgePredictionSequence(
                train_graph,
                batch_size=batch_size,
                batches_per_epoch=batches_per_epoch,
                negative_samples_rate=negative_samples_rate,
                use_edge_metrics=self._use_edge_metrics,
                sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types
            )
            if negative_valid_graph is not None:
                validation_sequence = EdgePredictionEvaluationSequence(
                    positive_graph=valid_graph,
                    negative_graph=negative_valid_graph,
                    batch_size=batch_size,
                    use_edge_metrics=self._use_edge_metrics,
                )
        elif self._task_name == "EDGE_LABEL_PREDICTION":
            training_sequence = EdgeLabelPredictionSequence(
                train_graph,
                batch_size=batch_size,
                batches_per_epoch=batches_per_epoch,
                use_edge_metrics=self._use_edge_metrics,
            )
            if valid_graph is not None:
                raise NotImplementedError(
                    "The validation sequence for edge label prediction "
                    "has not been implemented yet."
                )
        elif self._task_name == "BINARY_EDGE_LABEL_PREDICTION":
            training_sequence = BinaryEdgeLabelPredictionSequence(
                train_graph,
                positive_edge_type=self._positive_edge_type,
                batch_size=batch_size,
                batches_per_epoch=batches_per_epoch,
                use_edge_metrics=self._use_edge_metrics,
            )
            if valid_graph is not None:
                raise NotImplementedError(
                    "The validation sequence for binary edge label prediction "
                    "has not been implemented yet."
                )
        else:
            raise ValueError("Unreacheable!")

        return super().fit(
            training_sequence.into_dataset().repeat(),
            steps_per_epoch=training_sequence.steps_per_epoch,
            validation_data=None if validation_sequence is None else validation_sequence.into_dataset().repeat(),
            validation_steps=None if validation_sequence is None else validation_sequence.steps_per_epoch,
            epochs=epochs,
            early_stopping_monitor=early_stopping_monitor,
            early_stopping_min_delta=early_stopping_min_delta,
            early_stopping_patience=early_stopping_patience,
            early_stopping_mode=early_stopping_mode,
            reduce_lr_monitor=reduce_lr_monitor,
            reduce_lr_min_delta=reduce_lr_min_delta,
            reduce_lr_patience=reduce_lr_patience,
            reduce_lr_mode=reduce_lr_mode,
            reduce_lr_factor=reduce_lr_factor,
            verbose=verbose,
            **kwargs
        )

    def predict(self, *args, **kwargs):
        """Run predict."""
        return self._model.predict(*args, **kwargs)

    def evaluate(
        self,
        positive_graph: Graph,
        negative_graph: Optional[Graph] = None,
        batch_size: int = 2**10,
    ) -> Dict[str, float]:
        """Run predict."""
        self._validate_fit_parameters(
            negative_valid_graph=negative_graph,
            valid_graph=positive_graph
        )
        validation_sequence = None
        if self._task_name == "EDGE_PREDICTION":
            if negative_graph is None:
                raise ValueError(
                    "The negative graph was not provided."
                )
            validation_sequence = EdgePredictionEvaluationSequence(
                positive_graph=positive_graph,
                negative_graph=negative_graph,
                batch_size=batch_size,
                use_edge_metrics=self._use_edge_metrics,
            ).into_dataset()
        elif self._task_name == "EDGE_LABEL_PREDICTION":
            raise NotImplementedError(
                "The validation sequence for edge label prediction "
                "has not been implemented yet."
            )
        elif self._task_name == "BINARY_EDGE_LABEL_PREDICTION":
            raise NotImplementedError(
                "The validation sequence for binary edge label prediction "
                "has not been implemented yet."
            )
        else:
            raise ValueError("Unreacheable!")

        return dict(zip(
            self._model.metrics_names,
            self._model.evaluate(
                validation_sequence,
                batch_size=batch_size,
                verbose=False,
            )
        ))



"""Submodule providing models for edge-label prediction."""
from .edge_label_prediction_sklearn import SklearnModelEdgeLabelPredictionAdapter

__all__ = [
    "SklearnModelEdgeLabelPredictionAdapter"
]


"""Submodule providing pipelines for node-label evaluation and prediction."""
from typing import Union, Callable, List, Optional, Dict, Generator
import pandas as pd
import numpy as np
from ensmallen import Graph
from time import time
from tqdm.auto import trange, tqdm
from ensmallen.datasets import get_dataset
from sklearn.base import ClassifierMixin
import copy
from ..node_label_prediction import SklearnModelNodeLabelPredictionAdapter
from ..utils import is_sklearn_classifier_model, get_sklearn_default_classifier, is_default_sklearn_classifier
from .compute_node_embedding import compute_node_embedding


def _node_embedding_depends_on_node_types(node_embedding_method_name: str) -> bool:
    """Returns whether a given node embedding method name depends on the graph node types.

    The implications of such a dependency are that we can either compute the node embedding
    once for each provided graph when the start the pipeline, or we need to compute it
    multiple times, once per each of the holdouts.

    Parameters
    -------------------------
    node_embedding_method_name: str
        The method to check for.
    """
    # TODO: actually implement this method when other methods such as Siamese will start getting used.
    return False


def _compute_node_features(
    graph: Graph,
    compute_node_type_dependent_features: bool,
    node_features: Union[str, pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]],
    embedding_kwargs: Optional[Dict] = None,
    embedding_fit_kwargs: Optional[Dict] = None,
) -> List[Union[str, pd.DataFrame, np.ndarray]]:
    """Computes and returns all possible node features.

    Parameters
    ---------------------
    graph: Graph
        The graph to compute the features for.
    compute_node_type_dependent_features: bool
        Whether we can currently compute the node features that are
        dependent on the node types.
    node_features: Union[str, pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]]
        The node features to compute.
    embedding_kwargs: Optional[Dict] = None
        Kwargs to forward to the copmputation of the embedding.
    embedding_fit_kwargs: Optional[Dict] = None
        Kwargs to forward to the fit portion of the embedding.
    """
    if embedding_kwargs is None:
        embedding_kwargs = {}

    if not isinstance(node_features, list):
        node_features = [node_features]
    return [
        (
            compute_node_embedding(
                graph=graph,
                node_embedding_method_name=node_feature,
                fit_kwargs=embedding_fit_kwargs,
                verbose=False,
                **embedding_kwargs
            )[0]
            if not compute_node_type_dependent_features or _node_embedding_depends_on_node_types(node_feature)
            else node_feature
        ) if isinstance(node_feature, str) else node_feature
        for node_feature in node_features
    ]


def _validate_and_align_node_feature(
    graph: Graph,
    node_features: Union[pd.DataFrame, np.ndarray],
) -> np.ndarray:
    """Verifies that the node features provided are compatible with this graph

    Parameters
    ---------------------
    graph: Graph
        The graph to check for.
    node_features: Union[pd.DataFrame, np.ndarray]
        The features to validate and align.
    """
    if not isinstance(node_features, (np.ndarray, pd.DataFrame)):
        raise ValueError(
            (
                "The provided node features are of type {node_features_type}, "
                "while we only currently support numpy arrays and pandas DataFrames. "
                "What behaviour were you expecting with this feature? "
                "Please do open an issue on Embiggen and let us know!"
            ).format(
                node_features_type=type(node_features)
            )
        )

    if graph.get_nodes_number() != node_features.shape[0]:
        raise ValueError(
            (
                "The provided node features have {rows_number} rows "
                "but the provided graph{graph_name} has {nodes_number} nodes. "
                "Maybe these features refer to another "
                "version of the graph or another graph "
                "entirely?"
            ).format(
                rows_number=node_features.shape[0],
                graph_name="" if graph.get_name().lower(
                ) == "graph" else " {}".format(graph.get_name()),
                nodes_number=graph.get_nodes_number()
            )
        )

    # If it is a dataframe we align it
    if isinstance(node_features, pd.DataFrame):
        return node_features.loc[graph.get_node_names()].to_numpy()

    # And if it is a numpy array we must believe that the user knows what
    # they are doing, as we cannot ensure alignment.
    return node_features


def _validate_and_align_node_features(
    graph: Graph,
    node_features: List[Union[pd.DataFrame, np.ndarray]],
) -> List[np.ndarray]:
    """Verifies that the node features provided are compatible with this graph

    Parameters
    ---------------------
    graph: Graph
        The graph to check for.
    node_features: List[Union[pd.DataFrame, np.ndarray]]
        The features to validate and align.
    """
    return [
        _validate_and_align_node_feature(graph, nfs)
        for nfs in node_features
    ]


def _get_node_label_classifier_model(model, **kwargs: Dict):
    """Returns a new node label classifier model that may be trained.

    Parameters
    ------------------------
    model
        The model describing what the new model should look like.
        This object can be:
        - A string with the name of the desired model.
        - An instantiated Sklearn model that will be cloned and wrapped.
    **kwargs: Dict
        The kwargs to provide to the model constructor.
    """
    # If the provided model is a string
    if isinstance(model, str):
        if is_default_sklearn_classifier(model):
            return SklearnModelNodeLabelPredictionAdapter(
                get_sklearn_default_classifier(
                    model,
                    **kwargs
                )
            )
        raise ValueError(
            "The provided model name {} is not available.".format(
                model
            )
        )

    # If the provide model is already an sklearn model,
    # we proceed to wrap it and we avoid to training the original
    # models multiple times by making a deep copy of it.
    if is_sklearn_classifier_model(model):
        if kwargs:
            raise ValueError(
                "Kwargs for instantating a model were provided but "
                "the provided model is already instantiated. It is not "
                "clear what to do with these arguments."
            )
        return SklearnModelNodeLabelPredictionAdapter(
            copy.deepcopy(model)
        )

    raise ValueError(
        "It is not clear what to do with the provided model object of type {}.".format(
            type(model)
        )
    )

# TODO! Add cache!


def _execute_node_label_prediction_holdout(
    graph: Graph,
    graph_name: str,
    classifier_instance,
    number_of_holdouts: int,
    holdout_number: int,
    train_size: float,
    holdout_type: str,
    use_stratification: bool,
    random_state: int,
    graph_node_features: List[Union[str, pd.DataFrame, np.ndarray]],
    embedding_kwargs: Dict,
    embedding_fit_kwargs: Dict,
    classifier_kwargs: Dict,
    classifier_fit_kwargs: Dict
) -> List[Dict]:
    """Execute an holdout."""
    # Split the graph using the requested holdout type.
    if holdout_type == "Monte Carlo":
        train_graph, test_graph = graph.get_node_label_holdout_graphs(
            train_size=train_size,
            use_stratification=use_stratification,
            random_state=random_state*holdout_number
        )
    elif holdout_type == "KFold":
        train_graph, test_graph = graph.get_node_label_kfold(
            k=number_of_holdouts,
            k_index=holdout_number,
            use_stratification=use_stratification,
            random_state=random_state*holdout_number
        )
    else:
        raise ValueError(
            "The provided holdout type is not supported. "
            "We currently only support Monte Carlo and Kfold."
        )

    # Compute the remaining node features that we could not
    # compute before because of their dependence on the node types,
    # which would have caused a bias in the evaluation.
    # Once the features are computed, we also run a validation
    # on them to make sure that indeed they are compatible with
    # the provided graph.
    holdout_node_features = _validate_and_align_node_features(
        train_graph,
        _compute_node_features(
            train_graph,
            compute_node_type_dependent_features=True,
            node_features=graph_node_features,
            embedding_kwargs=embedding_kwargs,
            embedding_fit_kwargs=embedding_fit_kwargs
        )
    )

    start_training = time()
    classifier_instance.fit(
        graph=train_graph,
        node_features=holdout_node_features,
        behaviour_for_unknown_node_labels="drop",
        aligned_node_mapping=True,
        random_state=random_state,
    )
    training_time = time() - start_training

    both_performance = []

    for evaluation_type, graph_to_evaluate in (
        ("train", train_graph),
        ("test", test_graph),
    ):
        start_evaluation = time()
        performance = classifier_instance.evaluate(
            graph=graph_to_evaluate,
            node_features=holdout_node_features,
            behaviour_for_unknown_node_labels="drop",
            aligned_node_mapping=True,
            random_state=random_state,
        )
        evaluation_time = time() - start_evaluation

        # Add more metadata to the evaluation report.
        performance["evaluation_type"] = evaluation_type
        performance["size"] = train_size
        performance["graph_name"] = graph_name
        performance["holdout_type"] = holdout_type
        performance["required_training_time"] = training_time
        performance["required_evaluation_time"] = evaluation_time

        both_performance.append(performance)

    return both_performance


def _get_graphs_iterator(
    graphs: Union[List[Graph], List[str], Graph, str],
    repositories: Union[str, List[str]],
    versions: Union[str, List[str]],
    graph_normalization_callback: Callable[[Graph], Graph],
) -> List[Graph]:
    """Returns iterator on given graph objects.

    Parameters
    ------------------
    graphs: Union[Graph, str, List[str], List[Graph]]
        The graph to run the embedding and node-label prediction on.
        If a string was provided, we will retrieve the graphs from Ensmallen's repositories.
        If a list was provided, we will iterate on all graphs.
    repositories: Union[str, List[str]]
        Repositorie(s) of the graphs to be retrieved.
        This only applies when the provided graph(s) are all strings,
        otherwise an exception will be raised.
    versions: Union[str, List[str]]
        Version(s) of the graphs to be retrieved.
        This only applies when the provided graph(s) are all strings,
        otherwise an exception will be raised.
    graph_normalization_callback: Callable[[Graph], Graph]
        Graph normalization procedure to call on graphs that have been loaded from
        the Ensmallen automatic retrieval.
    """
    if not isinstance(graphs, list):
        graphs = [graphs]

    number_of_graphs = len(graphs)

    if versions is not None:
        if any(not isinstance(graph, str) for graph in graphs):
            raise ValueError(
                "Graph versions were provided, but the graphs are not ",
                "graph names from Ensmallen's automatic retrieval."
            )
        if isinstance(versions, str):
            versions = [versions] * number_of_graphs
    else:
        versions = [None] * number_of_graphs

    if repositories is not None:
        if any(not isinstance(graph, str) for graph in graphs):
            raise ValueError(
                "Graph repositories were provided, but the graphs are not ",
                "graph names from Ensmallen's automatic retrieval."
            )
        if isinstance(repositories, str):
            repositories = [repositories]*len(graphs)
    else:
        repositories = [None] * number_of_graphs

    for graph, repository, version in tqdm(
        zip(graphs, repositories, versions),
        desc="Running node-label prediction evaluations",
        disable=number_of_graphs == 1,
        total=number_of_graphs,
        dynamic_ncols=True,
        leave=False
    ):
        if isinstance(graph, str):
            graph = get_dataset(
                graph,
                repository=repository,
                version=version
            )()
            if graph_normalization_callback is not None:
                yield graph_normalization_callback(graph)
        yield graph


# TODO! Add caching decorator!
def _run_node_label_prediction_on_graph(
    graph: Graph,
    models: Union[str, Callable, ClassifierMixin, List[Union[str, Callable, ClassifierMixin]]],
    node_features: Union[str, pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]],
    number_of_holdouts: int,
    train_size: float,
    holdout_type: str,
    use_stratification: bool,
    random_state: int,
    embedding_kwargs: Optional[Dict],
    embedding_fit_kwargs: Optional[Dict],
    classifier_kwargs: Optional[Dict],
    classifier_fit_kwargs: Optional[Dict],
) -> List[Dict]:
    """Run the experiment on the provided graph."""
    # We compute all the node features we can
    # that are not dependent on the node types and therefore
    # do not cause a bias while evaluating them.
    graph_node_features = _compute_node_features(
        graph,
        compute_node_type_dependent_features=False,
        node_features=node_features,
        embedding_kwargs=embedding_kwargs,
        embedding_fit_kwargs=embedding_fit_kwargs
    )

    if classifier_kwargs is None:
        classifier_kwargs = {}

    if classifier_fit_kwargs is None:
        classifier_fit_kwargs = {}

    if not isinstance(models, list):
        models = [models]

    # And for each graph we compute a k-fold holdout
    return [
        performance
        for holdout_number in trange(
            number_of_holdouts,
            desc=f"Node-label prediction holdouts on {graph.get_name()}",
            dynamic_ncols=True,
            leave=False
        )
        # Iterate over the provided models
        for model in models
        # We compute the training and test performance on this holdout.
        for performance in _execute_node_label_prediction_holdout(
            graph,
            graph_name=graph.get_name(),
            classifier_instance=_get_node_label_classifier_model(
                model,
                **classifier_kwargs
            ),
            number_of_holdouts=number_of_holdouts,
            holdout_number=holdout_number,
            train_size=train_size,
            holdout_type=holdout_type,
            use_stratification=use_stratification,
            random_state=random_state,
            graph_node_features=graph_node_features,
            embedding_kwargs=embedding_kwargs,
            embedding_fit_kwargs=embedding_fit_kwargs,
            classifier_kwargs=classifier_kwargs,
            classifier_fit_kwargs=classifier_fit_kwargs
        )
    ]

# TODO! Add cache decorator!


def node_label_prediction(
    node_features: Union[str, pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]],
    graphs: Union[Graph, str, List[str], List[Graph]],
    models: Union[str, Callable, ClassifierMixin, List[Union[str, Callable, ClassifierMixin]]] = "DecisionTreeClassifier",
    repositories: Optional[Union[str, List[str]]] = None,
    versions: Optional[Union[str, List[str]]] = None,
    number_of_holdouts: int = 10,
    train_size: float = 0.8,
    holdout_type: str = "Monte Carlo",
    use_stratification: bool = True,
    random_state: int = 42,
    embedding_kwargs: Optional[Dict] = None,
    embedding_fit_kwargs: Optional[Dict] = None,
    classifier_kwargs: Optional[Dict] = None,
    classifier_fit_kwargs: Optional[Dict] = None,
    graph_normalization_callback: Callable[[Graph], Graph] = None,
) -> pd.DataFrame:
    """Return the evaluation of an embedding for edge prediction on the given model.

    Parameters
    ----------------------
    node_features: Union[str, pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]]
        The node features to be used to train the classifier.
        If a string is provided, the curresponding embedding will be computed.
        The features will be provided to the model always as a list.
    graphs: Union[Graph, str, List[str], List[Graph]]
        The graph to run the embedding and node-label prediction on.
        If a string was provided, we will retrieve the graphs from Ensmallen's repositories.
        If a list was provided, we will iterate on all graphs.
    model: Union[str, Callable, ClassifierMixin]
        Either the name of the model or a method returning a model.
    repositories: Optional[Union[str, List[str]]] = None
        Repositorie(s) of the graphs to be retrieved.
        This only applies when the provided graph(s) are all strings,
        otherwise an exception will be raised.
    versions: Optional[Union[str, List[str]]] = None
        Version(s) of the graphs to be retrieved.
        This only applies when the provided graph(s) are all strings,
        otherwise an exception will be raised.
    number_of_holdouts: int = 10
        The number of the holdouts to run.
    train_size: float = 0.8,
        Split size of the training graph.
        This value will be ignored if the holdouts is
        specified to be executed as a kfold.
    holdout_type: str = "Monte Carlo"
        Whether to use a Monte Carlo holdout or a Kfold
        holdout. The possible values are:
        - "Monte Carlo": holdout with repeated sampling, the different holdouts share the same complete set of samples.
        - "K fold": splits the dataset into k folds.
    use_stratification: bool = True
        Whether to use stratification.
        It is generally a VERY POOR CHOICE when running
        experiments to not use stratification as it may
        lead to biases dependending on your dataset labels
        distribution.
    random_state: int = 42
        The seed to be used to reproduce the holdout.
    embedding_fit_kwargs: Optional[Dict] = None
        The kwargs to be forwarded to the embedding fit method
    embedding_kwargs: Optional[Dict] = None
        The kwargs to be forwarded to the embedding method
    graph_normalization_callback: Callable[[Graph], Graph] = None
        Graph normalization procedure to call on graphs that have been loaded from
        the Ensmallen automatic retrieval.
    """
    return pd.DataFrame([
        holdout
        # We iterate over the provided graphs
        for graph in _get_graphs_iterator(
            graphs=graphs,
            repositories=repositories,
            versions=versions,
            graph_normalization_callback=graph_normalization_callback
        )
        for holdout in _run_node_label_prediction_on_graph(
            graph=graph,
            models=models,
            node_features=node_features,
            number_of_holdouts=number_of_holdouts,
            train_size=train_size,
            holdout_type=holdout_type,
            use_stratification=use_stratification,
            random_state=random_state,
            embedding_kwargs=embedding_kwargs,
            embedding_fit_kwargs=embedding_fit_kwargs,
            classifier_kwargs=classifier_kwargs,
            classifier_fit_kwargs=classifier_fit_kwargs,
        )
    ])



"""Submodule providing models for edge prediction."""
from .edge_prediction_sklearn import SklearnModelEdgePredictionAdapter


try:
    from .edge_prediction_tensorflow import (
        Perceptron, MultiLayerPerceptron, EdgePredictionGraphNeuralNetwork,
        FeedForwardNeuralNetwork, EdgePredictionModel
    )

    tensorflow_edge_prediction_models = {
        "perceptron": Perceptron,
        "multilayerperceptron": MultiLayerPerceptron,
    }

    def is_tensorflow_edge_prediction_method(model_name: str) -> bool:
        return model_name.lower() in tensorflow_edge_prediction_models

    def get_tensorflow_model(model_name: str, **kwargs) -> EdgePredictionModel:
        return tensorflow_edge_prediction_models[model_name.lower()](**kwargs)

    __all__ = [
        "SklearnModelEdgePredictionAdapter",
        "Perceptron",
        "MultiLayerPerceptron",
        "EdgePredictionGraphNeuralNetwork",
        "FeedForwardNeuralNetwork",
        "get_tensorflow_model"
    ]
except ModuleNotFoundError as e:
    def is_tensorflow_edge_prediction_method(model_name: str) -> bool:
        return False

    def get_tensorflow_model(model_name: str, **kwargs):
        raise NotImplementedError(
            "The method is not available when TensorFlow is not installed."
        )

    __all__ = [
        "SklearnModelEdgePredictionAdapter",
        "get_tensorflow_model"
    ]



"""Submodule providing pipelines for edge prediction."""
import itertools
from typing import Union, Callable, Tuple, List, Optional, Dict
import pandas as pd
from ensmallen import Graph
from time import time
from tqdm.auto import trange, tqdm
import math
from ensmallen.datasets import get_dataset
from sklearn.base import ClassifierMixin
import copy
import warnings
from ..edge_prediction import SklearnModelEdgePredictionAdapter, is_tensorflow_edge_prediction_method, get_tensorflow_model
from ..utils import is_sklearn_classifier_model, get_sklearn_default_classifier, is_default_sklearn_classifier
from .compute_node_embedding import compute_node_embedding


def get_negative_graphs(
    graph: Graph,
    random_state: int,
    sample_only_edges_with_heterogeneous_node_types: bool,
    unbalance_rate: float,
    train_size: float,
) -> Tuple[Graph, Graph]:
    """Return tuple with training and test negative graphs.

    Implementative details
    ----------------------------
    We generate first one larger graph of negative edges,
    and then we split it into a training and test subgraphs
    using the `random hodout` schema, that is we do not preserve
    the number of connected components as we do not care about them
    since we are not using the topology. Do note that this may
    not be the same case when using a GCN for edge prediction.

    Parameters
    ----------------------------
    graph: Graph
        The graph from which to sample the negative edges.
    random_state: int
        Random state to sample the negatives and execute the random split.
    sample_only_edges_with_heterogeneous_node_types: bool
        Whether to only sample edges between heterogeneous node types.
        This may be useful when training a model to predict between
        two portions in a bipartite graph.
    unbalance_rate: float
        Quantity over over (or under) sampling for the negatives respectively
        to the number of edges in the provided graph.
    train_size: float
        Split size of the training graph.
    """
    # For both the training and the test set we sample the same
    # number of negative edges are there are existing edges in the training and test graphs, respectively.
    # This is done in order to avoid an excessive bias in the evaluation of the edge prediction.
    # Across multiple holdouts, statistically it is unlikely to sample
    # consistently unknown positive edges, and therefore we should be able
    # to remove this negative bias from the evaluation.
    # Of course, this only apply to graphs where we can assume that there is
    # not a massive amount of unknown positive edges.
    negative_graph = graph.sample_negatives(
        number_of_negative_samples=int(
            math.ceil(graph.get_edges_number()*unbalance_rate)),
        random_state=random_state,
        sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
        verbose=False
    )
    return negative_graph.random_holdout(
        train_size=train_size,
        random_state=random_state,
        verbose=False,
    )


def evaluate_embedding_for_edge_prediction(
    embedding_method: Union[str, Callable[[Graph, int], pd.DataFrame]],
    graphs: Union[Graph, str, List[str], List[Graph]],
    model: Union[str, Callable, ClassifierMixin],
    repositories: Optional[Union[str, List[str]]] = None,
    versions: Optional[Union[str, List[str]]] = None,
    number_of_holdouts: int = 10,
    train_size: float = 0.8,
    unbalance_rates: Tuple[Union[float, str]] = (10.0, 100.0, "auto"),
    holdouts_random_state: int = 42,
    edge_embedding_method: str = "Concatenate",
    edge_types: Optional[List[str]] = None,
    minimum_node_degree: Optional[int] = None,
    maximum_node_degree: Optional[int] = None,
    use_only_cpu: Union[bool, str] = "auto",
    only_execute_embeddings: bool = False,
    embedding_kwargs: Optional[Dict] = None,
    embedding_fit_kwargs: Optional[Dict] = None,
    classifier_kwargs: Optional[Dict] = None,
    classifier_fit_kwargs: Optional[Dict] = None,
    subgraph_of_interest_for_edge_prediction: Optional[Graph] = None,
    sample_only_edges_with_heterogeneous_node_types: bool = False,
    graph_normalization_callback: Callable[[Graph], Graph] = None,
) -> pd.DataFrame:
    """Return the evaluation of an embedding for edge prediction on the given model.

    Parameters
    ----------------------
    embedding_method: Union[str, Callable[[Graph, int], pd.DataFrame]]
        Either the name of the embedding method or a method to be called.
    graphs: Union[Graph, str, List[str], List[Graph]]
        The graph to run the embedding and edge prediction on.
        If a string was provided, we will retrieve the graphs from Ensmallen's repositories.
        If a list was provided, we will iterate on all graphs.
    model: Union[str, Callable, ClassifierMixin]
        Either the name of the model or a method returning a model.
    repositories: Optional[Union[str, List[str]]] = None
        Repositorie(s) of the graphs to be retrieved.
        This only applies when the provided graph(s) are all strings,
        otherwise an exception will be raised.
    versions: Optional[Union[str, List[str]]] = None
        Version(s) of the graphs to be retrieved.
        This only applies when the provided graph(s) are all strings,
        otherwise an exception will be raised.
    number_of_holdouts: int = 10
        The number of the holdouts to run.
    train_size: float = 0.8
        Split size of the training graph.
    unbalance_rates: Tuple[Union[float, str]] = (10.0, 100.0, "auto")
        List of unbalance to evaluate.
        Do note that an unbalance of one is always included.
        With "auto", we use the true unbalance of the graph.
    holdouts_random_state: int = 42
        The seed to be used to reproduce the holdout.
    edge_types: Optional[List[str]] = None
        Edge types to focus the edge prediction on, if any.
    minimum_node_degree: Optional[int] = None
        The minimum node degree of either the source or destination node to be sampled. By default 0.
    maximum_node_degree: Optional[int] = None
        The maximum node degree of either the source or destination node to be sampled. By default, the number of nodes.
    use_only_cpu: Union[bool, str] = "auto",
        Whether to only use CPU.
        Do note that for CBOW and SkipGram models,
        this will switch the implementation from the
        TensorFlow implementation and will use our Rust Ensmallen one.
    only_execute_embeddings: bool = False
        Whether to only execute the computation of the embedding or also the edge prediction.
        This flag can be useful when the two operations should be executed on different machines
        or at different times.
    embedding_fit_kwargs: Optional[Dict] = None
        The kwargs to be forwarded to the embedding fit method
    embedding_kwargs: Optional[Dict] = None
        The kwargs to be forwarded to the embedding method
    subgraph_of_interest_for_edge_prediction: Optional[Graph] = None
        The subgraph to use for the edge prediction training and evaluation, if any.
        If none are provided, we sample the negative edges from the entire graph.
    sample_only_edges_with_heterogeneous_node_types: bool = False
        Whether to only sample edges between heterogeneous node types.
        This may be useful when training a model to predict between
        two portions in a bipartite graph.
    graph_normalization_callback: Callable[[Graph], Graph] = None
        Graph normalization procedure to call on graphs that have been loaded from
        the Ensmallen automatic retrieval.
    """

    if embedding_kwargs is None:
        embedding_kwargs = {}

    if classifier_kwargs is None:
        classifier_kwargs = {}

    if classifier_fit_kwargs is None:
        classifier_fit_kwargs = {}

    if edge_types is not None and isinstance(edge_types, str):
        edge_types = [edge_types]

    if isinstance(model, str):
        model_name = model
    elif is_sklearn_classifier_model(model):
        model_name = model.__class__.__name__
    elif callable(model):
        model_name = model.__name__

    if subgraph_of_interest_for_edge_prediction is not None and not subgraph_of_interest_for_edge_prediction.has_edges():
        raise ValueError(
            "The provided subgraph of interest does not have any edge."
        )

    if isinstance(graphs, (Graph, str)):
        graphs = [graphs]

    number_of_graphs = len(graphs)

    if versions is not None:
        if any(not isinstance(graph, str) for graph in graphs):
            raise ValueError(
                "Graph versions were provided, but the graphs are not ",
                "graph names from Ensmallen's automatic retrieval."
            )
        if isinstance(versions, str):
            versions = [versions] * number_of_graphs
    else:
        versions = [None] *number_of_graphs

    if repositories is not None:
        if any(not isinstance(graph, str) for graph in graphs):
            raise ValueError(
                "Graph repositories were provided, but the graphs are not ",
                "graph names from Ensmallen's automatic retrieval."
            )
        if isinstance(repositories, str):
            repositories = [repositories]*len(graphs)
    else:
        repositories = [None] *number_of_graphs

    holdouts = []
    for graph, repository, version in tqdm(
        zip(graphs, repositories, versions),
        desc="Executing graph",
        disable=number_of_graphs==1,
        dynamic_ncols=True,
        leave=False
    ):
        if isinstance(graph, str):
            graph = get_dataset(
                graph,
                repository=repository,
                version=version
            )()
            if graph_normalization_callback is not None:
                graph = graph_normalization_callback(graph)

        graph_name = graph.get_name()
        graph_unbalance_rate =  0.9 * (graph.get_nodes_number() * (graph.get_nodes_number() - 1)) / graph.get_number_of_directed_edges()

        unbalance_rates_for_graph = [
            graph_unbalance_rate
            if unbalance_rate == "auto"
            else unbalance_rate
            for unbalance_rate in unbalance_rates
            if unbalance_rate != "auto" and unbalance_rate < graph_unbalance_rate
        ]

        if len(unbalance_rates_for_graph) < len(unbalance_rates):
            warnings.warn(
                (
                    "Be advised that the provided unbalance rates included "
                    "rates that were higher than the maximum possible unbalance rate "
                    "possible in this graph {:4f}."
                ).format(
                    graph_unbalance_rate
                )
            )

        for holdout_number in trange(
            number_of_holdouts,
            desc=f"Executing holdouts on {graph_name}",
            dynamic_ncols=True,
            leave=False
        ):
            connected_holdout_random_state = holdouts_random_state*holdout_number
            connected_holdouts_parameters = dict(
                random_state=connected_holdout_random_state,
                edge_types=edge_types,
                minimum_node_degree=minimum_node_degree,
                maximum_node_degree=maximum_node_degree,
            )
            negative_graph_parameters = dict(
                sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
            )
            train_graph, test_graph = graph.connected_holdout(
                train_size,
                **connected_holdouts_parameters,
                verbose=False
            )

            start_embedding = time()
            if isinstance(embedding_method, str):
                embedding, _ = compute_node_embedding(
                    graph=train_graph,
                    node_embedding_method_name=embedding_method,
                    use_only_cpu=use_only_cpu,
                    fit_kwargs=embedding_fit_kwargs,
                    verbose=False,
                    **embedding_kwargs
                )
            else:
                embedding = embedding_method(
                    train_graph,
                    holdout_number,
                    **embedding_kwargs
                )
            seconds_necessary_for_embedding = time() - start_embedding

            if only_execute_embeddings:
                continue

            # If requested, we focus the training and test graphs
            # into the area of interest.
            if subgraph_of_interest_for_edge_prediction is not None:
                # We remove all the nodes from the training and test graphs
                # which are not part of the subgraph of interest.
                subgraph_node_names = subgraph_of_interest_for_edge_prediction.get_node_names()
                train_graph = train_graph.filter_from_names(
                    node_names_to_keep=subgraph_node_names
                )
                test_graph = test_graph.filter_from_names(
                    node_names_to_keep=subgraph_node_names
                )

                # We remove all the edges from the training and test graphs
                # which are not part of the subgraph of interest.
                train_graph = train_graph & subgraph_of_interest_for_edge_prediction
                test_graph = test_graph & subgraph_of_interest_for_edge_prediction

                assert train_graph.has_edges()
                assert test_graph.has_edges()

                # We realign the considered training and test graph with
                # the provided subgraph of interest node names dictionary.
                train_graph = train_graph.remap_from_graph(
                    subgraph_of_interest_for_edge_prediction
                )
                test_graph = test_graph.remap_from_graph(
                    subgraph_of_interest_for_edge_prediction
                )

                assert train_graph.has_edges()
                assert test_graph.has_edges()

            graph_to_use_to_sample_negatives = (
                # We sample the negative edges from the entire graph
                graph
                # when no subgraph of interest has been provided
                if subgraph_of_interest_for_edge_prediction is None
                # else, if the subgraph was provided, we only extract the negative
                # edges from this portion of the graph, making the evaluation of the edge prediction task
                # more significant for the actual desired task: often, when doing an edge predition
                # in a graph, we actually intend to predict some type of edges of interest.
                # When this is the case, we do not care to train or evaluate the model on edges that
                # are not in the portion of the graph.
                # Consider a graph representing a social network: if we are interested in learning the
                # connections between users living in small towns, if we also take into account users
                # living in large metropolis we would both train the model on unrelevant data and evaluate
                # the model of a different task, which may be more or less difficult.
                else subgraph_of_interest_for_edge_prediction
            )

            # Force alignment
            embedding = embedding.loc[train_graph.get_node_names()]

            train_graph.enable(
                vector_sources=True,
                vector_destinations=True,
                vector_cumulative_node_degrees=True
            )

            train_negative_graph, test_negative_graph = get_negative_graphs(
                graph_to_use_to_sample_negatives,
                # Inside ensmallen we pass this number in a splitmix function
                # therefore even a sum should have quite enough additional entropy.
                random_state=holdouts_random_state + holdout_number,
                **negative_graph_parameters,
                unbalance_rate=1.0,
                train_size=train_size,
            )

            if isinstance(model, str):
                if is_tensorflow_edge_prediction_method(model):
                    model_instance = get_tensorflow_model(
                        model,
                        graph=graph,
                        embedding=embedding,
                        edge_embedding_method=edge_embedding_method,
                        **classifier_kwargs
                    )
                elif is_default_sklearn_classifier(model):
                    model_instance = SklearnModelEdgePredictionAdapter(
                        get_sklearn_default_classifier(
                            model, **classifier_kwargs)
                    )
                else:
                    raise ValueError(
                        "The provided model name {} is not available.".format(
                            model)
                    )
            elif callable(model):
                if is_sklearn_classifier_model(model):
                    model_instance = SklearnModelEdgePredictionAdapter(
                        model(**classifier_kwargs))
                else:
                    model_instance = model(graph, embedding)
            elif is_sklearn_classifier_model(model):
                # If the provide model is already an sklearn model,
                # we proceed to wrap it and we avoid to training the original
                # models multiple times by making a deep copy of it.
                model_instance = SklearnModelEdgePredictionAdapter(
                    copy.deepcopy(model)
                )
            else:
                raise ValueError(
                    "It is not clear what to do with the provided model object of type {}.".format(
                        type(model)
                    )
                )

            start_training = time()
            if isinstance(model_instance, SklearnModelEdgePredictionAdapter):
                evaluation_kwargs = dict(
                    node_features=embedding.values,
                    edge_embedding_method=edge_embedding_method,
                    aligned_node_mapping=True,
                )
                model_instance.fit(
                    positive_graph=train_graph,
                    negative_graph=train_negative_graph,
                    # TODO: In the future consider adding these.
                    # edge_features: Optional[np.ndarray] = None,
                    **evaluation_kwargs
                )
            elif is_tensorflow_edge_prediction_method(model):
                evaluation_kwargs = dict()
                model_instance.fit(
                    train_graph=train_graph,
                    valid_graph=test_graph,
                    negative_valid_graph=test_negative_graph,
                    **classifier_fit_kwargs,
                    sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
                    verbose=False
                )
            else:
                raise NotImplementedError(
                    "Unclear what to do with this model."
                )
            seconds_necessary_to_train_model = time() - start_training

            start_train_evaluation = time()
            train_performance = model_instance.evaluate(
                positive_graph=train_graph,
                negative_graph=train_negative_graph,
                **evaluation_kwargs
            )
            train_evaluation_time = time() - start_train_evaluation

            train_performance["evaluation_type"] = "train"
            train_performance["unbalance"] = 1.0
            train_performance["size"] = train_size
            train_performance["graph_name"] = graph_name
            train_performance["required_training_time"] = seconds_necessary_to_train_model
            train_performance["seconds_necessary_for_embedding"] = seconds_necessary_for_embedding
            train_performance["train_evaluation_time"] = train_evaluation_time

            start_test_evaluation = time()
            test_performance = model_instance.evaluate(
                positive_graph=test_graph,
                negative_graph=test_negative_graph,
                **evaluation_kwargs
            )
            test_evaluation_time = time() - start_test_evaluation

            test_performance["evaluation_type"] = "test"
            test_performance["unbalance"] = 1.0
            test_performance["size"] = 1.0 - train_size
            test_performance["graph_name"] = graph_name
            test_performance["seconds_necessary_for_embedding"] = seconds_necessary_for_embedding
            train_performance["test_evaluation_time"] = test_evaluation_time

            for parameter_name, parameter in itertools.chain(
                connected_holdouts_parameters.items(),
                negative_graph_parameters.items(),
            ):
                if parameter is not None:
                    train_performance[parameter_name] = parameter
                    test_performance[parameter_name] = parameter

            holdouts.append(train_performance)
            holdouts.append(test_performance)

            for i, unbalance_rate in enumerate(tqdm(
                unbalance_rates_for_graph,
                desc="Evaluating on datasets with different unbalance",
                dynamic_ncols=True,
                leave=False
            ), start=1):
                train_negative_graph, test_negative_graph = get_negative_graphs(
                    graph_to_use_to_sample_negatives,
                    # Inside ensmallen we pass this number in a splitmix function
                    # therefore even a sum should have quite enough additional entropy.
                    random_state=holdouts_random_state + holdout_number + i,
                    sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
                    unbalance_rate=unbalance_rate,
                    train_size=train_size,
                )

                start_test_evaluation = time()
                train_performance = model_instance.evaluate(
                    positive_graph=train_graph,
                    negative_graph=train_negative_graph,
                    **evaluation_kwargs
                )
                train_evaluation_time = time() - start_train_evaluation

                train_performance["evaluation_type"] = "train"
                train_performance["unbalance"] = unbalance_rate
                train_performance["size"] = train_size
                train_performance["graph_name"] = graph_name
                train_performance["required_training_time"] = seconds_necessary_to_train_model
                train_performance["seconds_necessary_for_embedding"] = seconds_necessary_for_embedding
                train_performance["train_evaluation_time"] = train_evaluation_time

                start_test_evaluation = time()
                test_performance = model_instance.evaluate(
                    positive_graph=test_graph,
                    negative_graph=test_negative_graph,
                    **evaluation_kwargs
                )
                test_evaluation_time = time() - start_test_evaluation

                test_performance["evaluation_type"] = "test"
                test_performance["unbalance"] = unbalance_rate
                test_performance["size"] = 1.0 - train_size
                test_performance["graph_name"] = graph_name
                test_performance["seconds_necessary_for_embedding"] = seconds_necessary_for_embedding
                test_performance["test_evaluation_time"] = test_evaluation_time

                for parameter_name, parameter in itertools.chain(
                    connected_holdouts_parameters.items(),
                    negative_graph_parameters.items(),
                ):
                    if parameter is not None:
                        train_performance[parameter_name] = parameter
                        test_performance[parameter_name] = parameter

                holdouts.append(train_performance)
                holdouts.append(test_performance)

    holdouts = pd.DataFrame(holdouts)
    if isinstance(embedding_method, str):
        holdouts["embedding_method"] = embedding_method
    holdouts["model_name"] = model_name

    return holdouts



"""Abstract Keras Sequence object for running models on huge datasets."""
from typing import Union, Tuple, Dict
from keras_mixed_sequence import MixedSequence, VectorSequence
import numpy as np
import tensorflow as tf
from ..utils import tensorflow_version_is_higher_or_equal_than


class GloveSequence(MixedSequence):
    """Abstract Keras Sequence object for running models on huge datasets."""

    def __init__(
        self,
        sources: np.ndarray,
        destinations: np.ndarray,
        frequencies: np.ndarray,
        batch_size: int,
        directed: bool = False,
        shuffle: bool = True,
        elapsed_epochs: int = 0,
        random_state: int = 42
    ):
        """Create new Sequence object.

        Parameters
        -----------------------------
        sources: np.ndarray,
            Source elements.
        destinations: np.ndarray,
            Destination elements.
        frequencies: np.ndarray,
            Frequencies to be predicted by GloVe model.
        batch_size: int,
            Number of nodes to include in a single batch.
        directed: bool = False,
            Whether to randomly provide left and right inputs simmetrically
            or not.
        shuffle: bool = True,
            Whether to shuffle the values at each epoch.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        random_state: int = 42,
            Random random_state to make the sequence reproducible.
        """
        self._directed = directed
        batches_per_epoch = max(
            sources.size // batch_size,
            1
        )
        self._current_index = 0
        self._batches_per_epoch = batches_per_epoch
        super().__init__(
            x=[
                VectorSequence(
                    sources,
                    batch_size=batch_size,
                    random_state=random_state,
                    elapsed_epochs=elapsed_epochs,
                    shuffle=shuffle
                ),
                VectorSequence(
                    destinations,
                    batch_size=batch_size,
                    random_state=random_state,
                    elapsed_epochs=elapsed_epochs,
                    shuffle=shuffle
                )
            ],
            y=VectorSequence(
                frequencies,
                batch_size=batch_size,
                random_state=random_state,
                elapsed_epochs=elapsed_epochs,
                shuffle=shuffle
            )
        )

    def __call__(self):
        """Return next batch using an infinite generator model."""
        self._current_index = (self._current_index + 1) % self._batches_per_epoch
        return self[self._current_index]

    def into_dataset(self) -> tf.data.Dataset:
        """Return dataset generated out of the current sequence instance.

        Implementative details
        ---------------------------------
        This method handles the conversion of this Keras Sequence into
        a TensorFlow dataset, also handling the proper dispatching according
        to what version of TensorFlow is installed in this system.

        Returns
        ----------------------------------
        Dataset to be used for the training of a model
        """

        #################################################################
        # Handling kernel creation when TensorFlow is a modern version. #
        #################################################################

        if tensorflow_version_is_higher_or_equal_than("2.5.0"):
            return tf.data.Dataset.from_generator(
                self,
                output_signature=(
                    (
                        tf.TensorSpec(
                            shape=(None, ),
                            dtype=tf.int32
                        ),
                        tf.TensorSpec(
                            shape=(None, ),
                            dtype=tf.int32
                        )
                    ),
                    tf.TensorSpec(
                        shape=(None, ),
                        dtype=tf.float64
                    )
                )
            )

        return tf.data.Dataset.from_generator(
            self,
            output_types=(
                (
                    tf.int32,
                    tf.int32,
                ),
                tf.float64
            ),
            output_shapes=(
                (
                    tf.TensorShape([None, ]),
                    tf.TensorShape([None, ])
                ),
                tf.TensorShape([None, ])
            )
        )

    def __getitem__(self, idx: int) -> Tuple[
        Union[np.ndarray, Dict],
        Union[np.ndarray, Dict]
    ]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int
            Index corresponding to batch to be rendered.

        Returns
        ---------------
        Return Tuple containing input and output batches.
        """
        (sources, destinations), frequencies = super().__getitem__(idx)
        if self._directed or np.random.randint(2, dtype=bool):
            return (((sources, destinations), frequencies,),)
        return (((destinations, sources), frequencies,),)



"""Module providing adapter class making edge-label prediction possible in sklearn models."""
from sklearn.base import ClassifierMixin
from typing import Type, List, Dict, Union, Optional, Tuple
import pandas as pd
import numpy as np
from ensmallen import Graph
from ...utils import must_be_an_sklearn_classifier_model, evaluate_sklearn_classifier
from ...transformers import EdgeLabelPredictionTransformer, GraphTransformer


class SklearnModelEdgeLabelPredictionAdapter:
    """Class wrapping Sklearn models for running ."""

    def __init__(
        self,
        model_instance: Type[ClassifierMixin],
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: Type[ClassifierMixin]
            The class instance to be adapted into edge-label prediction.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        must_be_an_sklearn_classifier_model(model_instance)
        self._model_instance = model_instance

    @staticmethod
    def _trasform_graphs_into_edge_embedding(
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        behaviour_for_unknown_edge_labels: Optional[str] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and edge types extracted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the training of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the graph.
        behaviour_for_unknown_edge_labels: Optional[str] = None
            Behaviour to be followed when encountering edges that do not
            have a known edge type. Possible values are:
            - drop: we drop these edges
            - keep: we keep these edges
            By default, we drop these edges.
            If the behaviour has not been specified and left to None,
            a warning will be raised to notify the user of this uncertainty.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and training.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        lpt = EdgeLabelPredictionTransformer(
            method=edge_embedding_method,
            aligned_node_mapping=aligned_node_mapping
        )

        lpt.fit(node_features)

        return lpt.transform(
            graph=graph,
            edge_features=edge_features,
            behaviour_for_unknown_edge_labels=behaviour_for_unknown_edge_labels,
            random_state=random_state
        )

    @staticmethod
    def _trasform_graph_into_edge_embedding(
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
    ) -> np.ndarray:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Graph
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the training of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the graph.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        gt = GraphTransformer(
            method=edge_embedding_method,
            aligned_node_mapping=aligned_node_mapping
        )

        gt.fit(node_features)

        return gt.transform(
            graph=graph,
            edge_features=edge_features
        )

    def fit(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        behaviour_for_unknown_edge_labels: Optional[str] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
        **kwargs
    ):
        """Execute fitting of the model.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and edge types extracted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the training of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the graph.
        behaviour_for_unknown_edge_labels: Optional[str] = None
            Behaviour to be followed when encountering edges that do not
            have a known edge type. Possible values are:
            - drop: we drop these edges
            - keep: we keep these edges
            By default, we drop these edges.
            If the behaviour has not been specified and left to None,
            a warning will be raised to notify the user of this uncertainty.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and training.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model fitting.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        self._model_instance.fit(
            *SklearnModelEdgeLabelPredictionAdapter._trasform_graphs_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                edge_features=edge_features,
                behaviour_for_unknown_edge_labels=behaviour_for_unknown_edge_labels,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
                random_state=random_state,
            ),
            **kwargs
        )

    def evaluate(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        behaviour_for_unknown_edge_labels: Optional[str] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and edge types extracted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the evaluation of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the graph.
        behaviour_for_unknown_edge_labels: Optional[str] = None
            Behaviour to be followed when encountering edges that do not
            have a known edge type. Possible values are:
            - drop: we drop these edges
            - keep: we keep these edges
            By default, we drop these edges.
            If the behaviour has not been specified and left to None,
            a warning will be raised to notify the user of this uncertainty.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and evaluation.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return evaluate_sklearn_classifier(
            self._model_instance,
            *SklearnModelEdgeLabelPredictionAdapter._trasform_graphs_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                edge_features=edge_features,
                behaviour_for_unknown_edge_labels=behaviour_for_unknown_edge_labels,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
                random_state=random_state,
            ),
            multiclass_or_multilabel=graph.get_edge_types_number() > 2,
            **kwargs
        )

    def predict_proba(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the evaluation of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict_proba(
            SklearnModelEdgeLabelPredictionAdapter._trasform_graph_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                edge_features=edge_features,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
            ),
            **kwargs
        )

    def predict(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray],
        edge_features: Optional[np.ndarray] = None,
        edge_embedding_method: str = "Concatenate",
        aligned_node_mapping: bool = False,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the edge-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray]
            The node features to be used in the evaluation of the model.
        edge_features: Optional[np.ndarray] = None
            Optional edge features to be used as input concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.
        edge_embedding_method: str = "Concatenate"
            The method to be used to compute the edge embedding
            starting from the provided node features.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict(
            SklearnModelEdgeLabelPredictionAdapter._trasform_graph_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                edge_features=edge_features,
                edge_embedding_method=edge_embedding_method,
                aligned_node_mapping=aligned_node_mapping,
            ),
            **kwargs
        )



"""Siamese network for node-embedding including optionally node types and edge types."""
from typing import Optional, Union

import numpy as np
import pandas as pd
import tensorflow as tf
from ensmallen import Graph
from tensorflow.keras.layers import Embedding  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.constraints import UnitNorm  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras import backend as K  # pylint: disable=import-error,no-name-in-module
from tensorflow.keras.optimizers import \
    Optimizer  # pylint: disable=import-error,no-name-in-module

from .siamese import Siamese


class SimplE(Siamese):
    """Siamese network for node-embedding including optionally node types and edge types."""

    def __init__(
        self,
        graph: Graph,
        embedding_size: int = 100,
        embedding: Union[np.ndarray, pd.DataFrame] = None,
        model_name: str = "SimplE",
        optimizer: Union[str, Optimizer] = None,
        use_gradient_centralization: str = "auto"
    ):
        """Create new sequence TensorFlowEmbedder model.

        Parameters
        -------------------------------------------
        vocabulary_size: int = None,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding_size: int = 100,
            Dimension of the embedding.
            If None, the seed embedding must be provided.
            It is not possible to provide both at once.
        embedding: Union[np.ndarray, pd.DataFrame] = None,
            The seed embedding to be used.
            Note that it is not possible to provide at once both
            the embedding and either the vocabulary size or the embedding size.
        model_name: str = "SimplE",
            Name of the model.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
        use_gradient_centralization: bool = True,
            Whether to wrap the provided optimizer into a normalized
            one that centralizes the gradient.
            It is automatically enabled if the current version of
            TensorFlow supports gradient transformers.
            More detail here: https://arxiv.org/pdf/2004.01461.pdf
        """
        super().__init__(
            graph=graph,
            use_node_types=False,
            use_edge_types=True,
            embedding_size=embedding_size,
            edge_type_embedding_size=embedding_size,
            embedding=embedding,
            model_name=model_name,
            optimizer=optimizer,
            use_gradient_centralization=use_gradient_centralization
        )

    def _build_output(
        self,
        source_node_embedding: tf.Tensor,
        destination_node_embedding: tf.Tensor,
        edge_type_embedding: Optional[tf.Tensor] = None,
        edge_types_input: Optional[tf.Tensor] = None,
    ):
        """Return output of the model."""
        reverse_edge_type_embedding = Embedding(
            input_dim=self._edge_types_number,
            output_dim=self._edge_type_embedding_size,
            input_length=1,
            name="reverse_edge_type_embedding_layer",
        )(edge_types_input)
        reverse_edge_type_embedding = UnitNorm(
            axis=-1
        )(reverse_edge_type_embedding)
        return K.sum(
            source_node_embedding * edge_type_embedding * destination_node_embedding,
            axis=-1
        ) + K.sum(
            destination_node_embedding * reverse_edge_type_embedding * source_node_embedding,
            axis=-1
        )



"""Module providing abstract Node2Vec implementation."""
from typing import Optional
from ensmallen import Graph
import numpy as np
from ensmallen import models
from .ensmallen_embedder import EnsmallenEmbedder


class WeightedSPINE(EnsmallenEmbedder):
    """Abstract class for Node2Vec algorithms."""

    def __init__(
        self,
        embedding_size: int = 100,
        use_edge_weights_as_probabilities: bool = False,
        verbose: bool = True
    ):
        """Create new abstract Node2Vec method.

        Parameters
        --------------------------
        embedding_size: int = 100
            Dimension of the embedding.
        use_edge_weights_as_probabilities: bool = False
            Whether to treat the weights as probabilities.
        verbose: bool = True
            Whether to show loading bars
        """
        self._WeightedSPINE = models.WeightedSPINE(
            embedding_size=embedding_size,
            use_edge_weights_as_probabilities=use_edge_weights_as_probabilities
        )

        super().__init__(
            embedding_size=embedding_size,
            verbose=verbose
        )

    def _fit_transform(self, graph: Graph) -> np.ndarray:
        return self._WeightedSPINE.fit_transform(
            graph,
            verbose=self._verbose,
        ).T



"""Module providing adapter class making node-label prediction possible in sklearn models."""
from sklearn.base import ClassifierMixin
from typing import Type, List, Dict, Union, Optional, Tuple
import pandas as pd
import numpy as np
from ensmallen import Graph
from ...transformers import NodeLabelPredictionTransformer, NodeTransformer
from ...utils import must_be_an_sklearn_classifier_model, evaluate_sklearn_classifier


class SklearnModelNodeLabelPredictionAdapter:
    """Class wrapping Sklearn models for running node-label predictions."""

    def __init__(
        self,
        model_instance: Type[ClassifierMixin],
    ):
        """Create the adapter for Sklearn object.

        Parameters
        ----------------
        model_instance: Type[ClassifierMixin]
            The class instance to be adapted into node-label prediction.

        Raises
        ----------------
        ValueError
            If the provided model_instance is not a subclass of `ClassifierMixin`.
        """
        must_be_an_sklearn_classifier_model(model_instance)
        self._model_instance = model_instance
        # We want to mask the decorator class name
        self.__class__.__name__ = model_instance.__class__.__name__
        self.__class__.__doc__ = model_instance.__class__.__doc__

    @staticmethod
    def _trasform_graphs_into_node_embedding(
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        behaviour_for_unknown_node_labels: str = "warn",
        aligned_node_mapping: bool = False,
        shuffle: bool = False,
        random_state: int = 42,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Graph,
            The graph to use for this task.
            It can either be an Graph or a list of lists of nodes.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            The node features to be used in the training of the model.
        behaviour_for_unknown_node_labels: str = "warn"
            Behaviour to be followed when encountering nodes that do not
            have a known node type. Possible values are:
            - drop: we drop these nodes
            - keep: we keep these nodes
            By default, we drop these nodes.
            If the behaviour has not been specified and left to None,
            a warning will be raised to notify the user of this uncertainty.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        shuffle: bool = False
            Whether to shuffle the labels.
            In some models, this is necessary.
        random_state: int = 42
            The random state to reproduce the sampling and training.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        nlpt = NodeLabelPredictionTransformer(
            aligned_node_mapping=aligned_node_mapping
        )

        nlpt.fit(node_features)

        return nlpt.transform(
            graph=graph,
            behaviour_for_unknown_node_labels=behaviour_for_unknown_node_labels,
            shuffle=shuffle,
            random_state=random_state
        )

    @staticmethod
    def _trasform_graph_into_node_embedding(
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        aligned_node_mapping: bool = False,
    ) -> np.ndarray:
        """Transforms the provided data into an Sklearn-compatible numpy array.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            The node features to be used in the training of the model.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        gt = NodeTransformer(aligned_node_mapping=aligned_node_mapping)
        gt.fit(node_features)
        return gt.transform(graph,)

    def fit(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        behaviour_for_unknown_node_labels: str = "warn",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
        **kwargs
    ):
        """Execute fitting of the model.

        Parameters
        ------------------
        graph: Graph,
            The graph to use for this task.
            It can either be an Graph or a list of lists of nodes.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            The node features to be used in the training of the model.
        behaviour_for_unknown_node_labels: str = "warn"
            Behaviour to be followed when encountering nodes that do not
            have a known node type. Possible values are:
            - drop: we drop these nodes
            - keep: we keep these nodes
            By default, we drop these nodes.
            If the behaviour has not been specified and left to None,
            a warning will be raised to notify the user of this uncertainty.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and training.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model fitting.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        self._model_instance.fit(
            *SklearnModelNodeLabelPredictionAdapter._trasform_graphs_into_node_embedding(
                graph=graph,
                node_features=node_features,
                behaviour_for_unknown_node_labels=behaviour_for_unknown_node_labels,
                aligned_node_mapping=aligned_node_mapping,
                random_state=random_state,
            ),
            **kwargs
        )

    def evaluate(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        behaviour_for_unknown_node_labels: str = "warn",
        aligned_node_mapping: bool = False,
        random_state: int = 42,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the node-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph to use for this task.
            It can either be an Graph or a list of lists of nodes.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            The node features to be used in the evaluation of the model.
        behaviour_for_unknown_node_labels: str = "warn"
            Behaviour to be followed when encountering nodes that do not
            have a known node type. Possible values are:
            - drop: we drop these nodes
            - keep: we keep these nodes
            By default, we drop these nodes.
            If the behaviour has not been specified and left to None,
            a warning will be raised to notify the user of this uncertainty.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42
            The random state to reproduce the sampling and evaluation.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return evaluate_sklearn_classifier(
            self._model_instance,
            *SklearnModelNodeLabelPredictionAdapter._trasform_graphs_into_node_embedding(
                graph=graph,
                node_features=node_features,
                behaviour_for_unknown_node_labels=behaviour_for_unknown_node_labels,
                aligned_node_mapping=aligned_node_mapping,
                shuffle=False,
                random_state=random_state,
            ),
            multiclass_or_multilabel=graph.get_node_types_number() > 2,
            **kwargs
        )

    def predict_proba(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        aligned_node_mapping: bool = False,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the node-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph whose edges are to be embedded and predicted.
            It can either be an Graph or a list of lists of edges.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            The node features to be used in the evaluation of the model.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict_proba(
            SklearnModelNodeLabelPredictionAdapter._trasform_graph_into_edge_embedding(
                graph=graph,
                node_features=node_features,
                aligned_node_mapping=aligned_node_mapping,
            ),
            **kwargs
        )

    def predict(
        self,
        graph: Graph,
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]],
        aligned_node_mapping: bool = False,
        **kwargs
    ) -> Dict[str, float]:
        """Return evaluations of the model on the node-label prediction task on the provided data.

        Parameters
        ------------------
        graph: Graph,
            The graph to use for this task.
            It can either be an Graph or a list of lists of nodes.
        node_features: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            The node features to be used in the evaluation of the model.
        aligned_node_mapping: bool = False,
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        **kwargs: Dict
            Dictionary of kwargs to be forwarded to sklearn model prediction.

        Raises
        ------------------
        ValueError
            If the two graphs do not share the same node vocabulary.
        """
        return self._model_instance.predict(
            SklearnModelNodeLabelPredictionAdapter._trasform_graph_into_node_embedding(
                graph=graph,
                node_features=node_features,
                aligned_node_mapping=aligned_node_mapping,
            ),
            **kwargs
        )



"""Keras Sequence object for running CBOW and SkipGram on texts."""
from typing import Tuple, List
import tensorflow as tf
from ..utils import tensorflow_version_is_higher_or_equal_than
import numpy as np  # type: ignore
from ensmallen import preprocessing  # pylint: disable=no-name-in-module
from keras_mixed_sequence import VectorSequence

from .abstract_sequence import AbstractSequence


class Word2VecSequence(AbstractSequence):
    """Keras Sequence object for running CBOW and SkipGram on texts."""

    def __init__(
        self,
        sequences: List[np.ndarray],
        batch_size: int,
        window_size: int = 4,
        random_state: int = 42,
        elapsed_epochs: int = 0,
    ):
        """Create new Node2Vec Sequence object.

        Parameters
        -----------------------------
        sequences: List[np.ndarray]
            List of sequences of integers.
        batch_size: int
            Number of nodes to include in a single batch.
        window_size: int = 4
            Window size for the local context.
            On the borders the window size is trimmed.
        random_state: int = 42
            The random_state to use to make extraction reproducible.
        elapsed_epochs: int = 0
            Number of elapsed epochs to init state of generator.
        """

        self._sequences = VectorSequence(
            sequences,
            batch_size,
            random_state=random_state,
            elapsed_epochs=elapsed_epochs
        )
        self._current_index = 0
        super().__init__(
            window_size=window_size,
            sample_number=self._sequences.sample_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs,
            random_state=random_state
        )

    def on_epoch_end(self):
        """Shuffles given sequences object."""
        super().on_epoch_end()
        self._sequences.on_epoch_end()

    def __call__(self):
        """Return next batch using an infinite generator model."""
        self._current_index = self._current_index % self._sequences.steps_per_epoch
        return self[self._current_index]

    def into_dataset(self) -> tf.data.Dataset:
        """Return dataset generated out of the current sequence instance.

        Implementative details
        ---------------------------------
        This method handles the conversion of this Keras Sequence into
        a TensorFlow dataset, also handling the proper dispatching according
        to what version of TensorFlow is installed in this system.

        Returns
        ----------------------------------
        Dataset to be used for the training of a model
        """

        #################################################################
        # Handling kernel creation when TensorFlow is a modern version. #
        #################################################################

        if tensorflow_version_is_higher_or_equal_than("2.5.0"):
            return tf.data.Dataset.from_generator(
                self,
                output_signature=(
                    (
                        tf.TensorSpec(
                            shape=(None, self._window_size*2),
                            dtype=tf.int32
                        ),
                        tf.TensorSpec(
                            shape=(None, ),
                            dtype=tf.int32
                        )
                    ),
                )
            )

        return tf.data.Dataset.from_generator(
            self,
            output_types=(
                (
                    tf.int32,
                    tf.int32
                ),
            ),
            output_shapes=(
                (
                    tf.TensorShape([None, self._window_size*2]),
                    tf.TensorShape([None, ])
                ),
            )
        )

    def __getitem__(self, idx: int) -> Tuple[Tuple[np.ndarray, np.ndarray], None]:
        """Return batch corresponding to given index.

        The return tuple of tuples is composed of an inner tuple, containing
        the words vector and the vector of vectors of the contexts.
        Depending on the order of the input_layers of the models that can
        accept these data format, one of the vectors is used as training
        input and the other one is used as the output for the NCE loss layer.

        The words vectors and contexts vectors contain numeric IDs, that
        represent the index of the words' embedding column.

        The true output value is None, since no loss function is used after
        the NCE loss, that is implemented as a layer, and this vastly improves
        the speed of the training process since it does not require to allocate
        empty vectors of considerable size for the one-hot encoding process.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Tuple of tuples with input data.
        """
        contexts, words = preprocessing.word2vec(
            self._sequences[idx],
            window_size=self._window_size,
        )

        return (((contexts, words),),)



"""Submodule providing node embedding models implemented in Ensmallen in Rust.

These models are NOT dependant on TensorFlow and execute in CPU, not GPU.
"""

from .ensmallen_embedder import EnsmallenEmbedder
from .graph_cbow import GraphCBOW
from .graph_skipgram import GraphSkipGram
from .spine import SPINE
from .weighted_spine import WeightedSPINE

__all__ = [
    "GraphCBOW",
    "GraphSkipGram",
    "SPINE",
    "WeightedSPINE",
    "EnsmallenEmbedder"
]


"""Submodule providing TensorFlow and Ensmallen-based embedders."""
try:
    from . import tensorflow_embedders
    from . import ensmallen_embedders

    SUPPORTED_NODE_EMBEDDING_METHODS = {
        "cbow": {
            "gpu": tensorflow_embedders.GraphCBOW,
            "cpu": ensmallen_embedders.GraphCBOW,
        },
        "glove": tensorflow_embedders.GraphGloVe,
        "skipgram": {
            "gpu": tensorflow_embedders.GraphSkipGram,
            "cpu": ensmallen_embedders.GraphSkipGram,
        },
        "siamese": tensorflow_embedders.Siamese,
        "transe": tensorflow_embedders.TransE,
        "simple": tensorflow_embedders.SimplE,
        "transh": tensorflow_embedders.TransH,
        "transr": tensorflow_embedders.TransR,
        "spine": ensmallen_embedders.SPINE,
        "weightedspine": ensmallen_embedders.WeightedSPINE,
    }

    __all__ = [
        "tensorflow_embedders",
        "ensmallen_embedders",
        "SUPPORTED_NODE_EMBEDDING_METHODS"
    ]
except ModuleNotFoundError as e:
    from .ensmallen_embedders import *

    SUPPORTED_NODE_EMBEDDING_METHODS = {
        "cbow": GraphCBOW,
        "skipgram": GraphSkipGram,
        "spine": SPINE,
        "weightedspine": WeightedSPINE
    }

    __all__ = [
        "ensmallen_embedders",
        "SUPPORTED_NODE_EMBEDDING_METHODS",
    ]



"""Keras Sequence for running Neural Network on graph edge prediction."""
from typing import Tuple, Union, Optional
import warnings

import numpy as np
from ensmallen import Graph  # pylint: disable=no-name-in-module
from keras_mixed_sequence import Sequence
import tensorflow as tf
from ..utils import tensorflow_version_is_higher_or_equal_than


class EdgePredictionEvaluationSequence(Sequence):
    """Keras Sequence for running Neural Network on graph edge prediction."""

    def __init__(
        self,
        positive_graph: Graph,
        negative_graph: Optional[Graph] = None,
        use_node_types: bool = False,
        use_edge_types: bool = False,
        use_edge_metrics: bool = False,
        batch_size: int = 2**10,
        filter_none_values: bool = True,
        elapsed_epochs: int = 0,
        random_state: int = 42
    ):
        """Create new EdgePredictionSequence object.

        Parameters
        --------------------------------
        positive_graph: Graph
            The graph from which to sample the positive edges.
        negative_graph: Optional[Graph] = None
            The graph from which to sample the negative edges.
        use_node_types: bool = False
            Whether to return the node types.
        use_edge_types: bool = False
            Whether to return the edge types.
        use_edge_metrics: bool = False
            Whether to return the edge metrics.
        batch_size: int = 2**10
            The batch size to use.
        filter_none_values: bool = True
            Whether to filter None values.
        elapsed_epochs: int = 0
            Number of elapsed epochs to init state of generator.
        random_state: int = 42
            The random_state to use to make extraction reproducible.

        Raises
        --------------------------
        ValueError
            If the two graphs do not share the same vocabulary and therefore
            are not compatible with each other (or the model, likely).
        ValueError
            If the positive and negative graphs overlap.
        ValueError
            If either of the provided graphs is empty.
        """
        if negative_graph is not None:
            if not positive_graph.has_compatible_node_vocabularies(negative_graph):
                raise ValueError(
                    "The two provided positive and negative graphs are "
                    "not compatible with each other, for instance they may not "
                    "share the same dictionary of nodes."
                )
            if positive_graph.overlaps(negative_graph):
                raise ValueError(
                    "The provided positive and negative graphs overlap "
                    "with each other, making the definition of true negative "
                    "and true positive within this evaluation ambigous."
                )
            if not negative_graph.has_edges():
                raise ValueError(
                    "The provided negative graph is empty, that is, it has no edges."
                )
            if positive_graph.has_selfloops() != negative_graph.has_selfloops():
                warnings.warn(
                    "Please be advides that in either the provided positive "
                    "or negative graphs there are present selfloops. "
                    "If such odd topologies were also exclusively "
                    "present during the training of this model in only either the positive "
                    "or negative edges, the model might learn that self-loops are always "
                    "positive and/or negative. This might be okay or not according "
                    "to the experimental setup you have designed."
                )

        if not positive_graph.has_edges():
            raise ValueError(
                "The provided positive graph is empty, that is, it has no edges."
            )

        self._positive_graph = positive_graph
        self._negative_graph = negative_graph
        self._random_state = random_state
        self._use_node_types = use_node_types
        self._use_edge_types = use_edge_types
        self._filter_none_values = filter_none_values
        self._use_edge_metrics = use_edge_metrics
        edges_number = positive_graph.get_number_of_directed_edges()
        if negative_graph is not None:
            edges_number += negative_graph.get_number_of_directed_edges()
        batches_per_epoch = max(
            edges_number // batch_size,
            1
        )
        self._current_index = 0
        self._batches_per_epoch = batches_per_epoch
        super().__init__(
            sample_number=edges_number,
            batch_size=batch_size,
            elapsed_epochs=elapsed_epochs
        )

    def __call__(self):
        """Return next batch using an infinite generator model."""
        self._current_index = (self._current_index + 1) % self._batches_per_epoch
        return self[self._current_index]

    def into_dataset(self) -> tf.data.Dataset:
        """Return dataset generated out of the current sequence instance.

        Implementative details
        ---------------------------------
        This method handles the conversion of this Keras Sequence into
        a TensorFlow dataset, also handling the proper dispatching according
        to what version of TensorFlow is installed in this system.

        Returns
        ----------------------------------
        Dataset to be used for the training of a model
        """

        #################################################################
        # Handling kernel creation when TensorFlow is a modern version. #
        #################################################################

        if tensorflow_version_is_higher_or_equal_than("2.5.0"):
            input_tensor_specs = []

            for _ in range(2):
                # Shapes of the source and destination node IDs
                input_tensor_specs.append(tf.TensorSpec(
                    shape=(None, ), dtype=tf.int32))
                if self._use_node_types:
                    # Shapes of the source and destination node type IDs
                    input_tensor_specs.append(tf.TensorSpec(
                        shape=(None, self._graph.get_maximum_multilabel_count()), dtype=tf.int32))

            if self._use_edge_metrics:
                # Shapes of the edge type IDs
                input_tensor_specs.append(tf.TensorSpec(shape=(
                    None, self._graph.get_number_of_available_edge_metrics()), dtype=tf.float64))

            if self._use_edge_types:
                # Shapes of the edge type IDs
                input_tensor_specs.append(tf.TensorSpec(
                    shape=(None, ), dtype=tf.int32))

            return tf.data.Dataset.from_generator(
                self,
                output_signature=(
                    (
                        *input_tensor_specs,
                    ),
                    tf.TensorSpec(
                        shape=(None,),
                        dtype=tf.bool
                    )
                )
            )

        input_tensor_types = []
        input_tensor_shapes = []

        for _ in range(2):
            input_tensor_types.append(tf.int32,)
            input_tensor_shapes.append(tf.TensorShape([None, ]),)

            if self._use_node_types:
                input_tensor_types.append(tf.int32,)
                input_tensor_shapes.append(
                    tf.TensorShape([None, self._graph.get_maximum_multilabel_count()]),)

        if self._use_edge_metrics:
            input_tensor_types.append(tf.float64,)
            input_tensor_shapes.append(tf.TensorShape(
                [None, self._graph.get_number_of_available_edge_metrics()]),)

        if self._use_edge_types:
            input_tensor_types.append(tf.int32,)
            input_tensor_shapes.append(tf.TensorShape([None, ]),)

        return tf.data.Dataset.from_generator(
            self,
            output_types=(
                (
                    *input_tensor_types,
                ),
                tf.bool
            ),
            output_shapes=(
                (
                    *input_tensor_shapes,
                ),
                tf.TensorShape([None, ]),
            )
        )

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        """Return batch corresponding to given index.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Return Tuple containing X and Y numpy arrays corresponding to given batch index.
        """
        sources, source_node_types, destinations, destination_node_types, edge_metrics, edge_types = self._positive_graph.get_edge_prediction_chunk_mini_batch(
            idx,
            return_node_types=self._use_node_types,
            return_edge_types=False,
            return_edge_metrics=self._use_edge_metrics,
            batch_size=self.batch_size//2,
        )
        labels = np.ones_like(sources, dtype=bool)

        if self._negative_graph is not None:
            negative_sources, negative_source_node_types, negative_destinations, negative_destination_node_types, negative_edge_metrics, negative_edge_types = self._negative_graph.get_edge_prediction_chunk_mini_batch(
                idx,
                return_node_types=self._use_node_types,
                return_edge_types=False,
                return_edge_metrics=self._use_edge_metrics,
                batch_size=self.batch_size//2,
            )
            negative_labels = np.zeros_like(negative_sources, dtype=bool)

            sources = np.hstack([
                sources,
                negative_sources
            ])
            if source_node_types is not None:
                source_node_types = np.hstack([
                    source_node_types,
                    negative_source_node_types
                ])
            destinations = np.hstack([
                destinations,
                negative_destinations
            ])
            if destination_node_types is not None:
                destination_node_types = np.hstack([
                    destination_node_types,
                    negative_destination_node_types
                ])
            if edge_metrics is not None:
                edge_metrics = np.hstack([
                    edge_metrics,
                    negative_edge_metrics
                ])
            if edge_types is not None:
                edge_types = np.hstack([
                    edge_types,
                    negative_edge_types
                ])
            labels = np.hstack([
                labels,
                negative_labels
            ])

        return ((tuple([
            value
            for value in (
                sources, source_node_types, destinations, destination_node_types, edge_metrics, edge_types,
            )
            if not self._filter_none_values or value is not None
        ]), labels,), )
